#!/usr/bin/env python3
# vim: fileencoding=utf-8 ts=4
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: © Copyright 2025 by Christian Dönges. All rights reserved.
# SPDXID: SPDXRef-test-utilities-convert-to-si

from dataclasses import dataclass
import math

import pytest
from pymisclib.prefixes import convert_to_prefix, UnitPrefix


@dataclass
class TestVector:
    """Test vector to test conversion to SI-prefixed."""
    value: float
    """Value to convert."""
    expected_scaled: float
    """Expected scaled value."""
    expected_prefix: str
    """Expected metric prefix."""


test_vectors_binary_selected = [
    TestVector(0, 0, '  '),  # trivial
    TestVector(1, 1, '  '),  # also trivial
    # Progression of values.
    TestVector(1.073741824, 1.073741824, '  '),
    TestVector(10.73741824, 10.73741824, '  '),
    TestVector(107.3741824, 107.3741824, '  '),
    TestVector(1073.741824, 1.048576, 'Ki'),
    TestVector(10737.41824, 10.48576, 'Ki'),
    TestVector(107374.1824, 104.8576, 'Ki'),
    TestVector(1073741.824, 1.024, 'Mi'),
    TestVector(10737418.24, 10.24, 'Mi'),
    TestVector(107374182.4, 102.4, 'Mi'),
    TestVector(1073741824, 1, 'Gi'),
    # Test around the switch from Mi to Gi
    TestVector(1073741823, 1023.999999, 'Mi'),
    TestVector(1073741824, 1, 'Gi'),
    TestVector(1073741825, 1, 'Gi'),
]  # test_vectors_binary_selected

test_vectors_metric_selected = [
    TestVector(0, 0, ' '),  # trivial test
    TestVector(1, 1, ' '),  # another trivial test
    # test a progression of values
    TestVector(1.2356494, 1.2356494, ' '),
    TestVector(12.356494, 12.356494, ' '),
    TestVector(123.56494, 123.56494, ' '),
    TestVector(1235.6494, 1.2356494, 'k'),
    TestVector(12356.494, 12.356494, 'k'),
    TestVector(123564.94, 123.56494, 'k'),
    TestVector(1235649.4, 1.2356494, 'M'),
    # test around the switch from M to G
    TestVector(999999999, 999.999999, 'M'),
    TestVector(1000000000, 1.0, 'G'),
    TestVector(1000000001, 1.000000001, 'G'),
    # Test progression of small values
    TestVector(0.987654321, 987.654321, 'm'),
    TestVector(0.0987654321, 98.7654321, 'm'),
    TestVector(0.00987654321, 9.87654321, 'm'),
    TestVector(0.000987654321, 987.654321, 'µ'),
    TestVector(0.0000987654321, 98.7654321, 'µ'),
    TestVector(0.00000987654321, 9.87654321, 'µ'),
]  # test_vectors_metric_selected

test_vectors_binary_all_exponents = [
    TestVector(0, 0, '  '),
    TestVector(1024, 1, 'Ki'),
    TestVector(1024*1024, 1, 'Mi'),
    TestVector(1024*1024*1024, 1, 'Gi'),
    TestVector(1024*1024*1024*1024, 1, 'Ti'),
    TestVector(1024*1024*1024*1024*1024, 1, 'Pi'),
    TestVector(1024*1024*1024*1024*1024*1024, 1, 'Ei'),
    TestVector(1024*1024*1024*1024*1024*1024*1024, 1, 'Zi'),
    TestVector(1024*1024*1024*1024*1024*1024*1024*1024, 1, 'Yi'),
    TestVector(1024*1024*1024*1024*1024*1024*1024*1024*1024, 1, 'Ri'),
    TestVector(1024*1024*1024*1024*1024*1024*1024*1024*1024*1024, 1, 'Qi'),
]  # test_vectors_binary_all_exponents

test_vectors_metric_all_exponents = [
    TestVector(1e-24, 1, 'y'),
    TestVector(1e-23, 10, 'y'),
    TestVector(1e-22, 100, 'y'),
    TestVector(1e-21, 1, 'z'),
    TestVector(1e-20, 10, 'z'),
    TestVector(1e-19, 100, 'z'),
    TestVector(1e-18, 1, 'a'),
    TestVector(1e-17, 10, 'a'),
    TestVector(1e-16, 100, 'a'),
    TestVector(1e-15, 1, 'f'),
    TestVector(1e-14, 10, 'f'),
    TestVector(1e-13, 100, 'f'),
    TestVector(1e-12, 1, 'p'),
    TestVector(1e-11, 10, 'p'),
    TestVector(1e-10, 100, 'p'),
    TestVector(1e-9, 1, 'n'),
    TestVector(1e-8, 10, 'n'),
    TestVector(1e-7, 100, 'n'),
    TestVector(1e-6, 1, 'µ'),
    TestVector(1e-5, 10, 'µ'),
    TestVector(1e-4, 100, 'µ'),
    TestVector(1e-3, 1, 'm'),
    TestVector(1e-2, 10, 'm'),
    TestVector(1e-1, 100, 'm'),
    TestVector(1, 1, ' '),
    TestVector(1e1, 10, ' '),
    TestVector(1e2, 100, ' '),
    TestVector(1e3, 1, 'k'),
    TestVector(1e4, 10, 'k'),
    TestVector(1e5, 100, 'k'),
    TestVector(1e6, 1, 'M'),
    TestVector(1e7, 10, 'M'),
    TestVector(1e8, 100, 'M'),
    TestVector(1e9, 1, 'G'),
    TestVector(1e10, 10, 'G'),
    TestVector(1e11, 100, 'G'),
    TestVector(1e12, 1, 'T'),
    TestVector(1e13, 10, 'T'),
    TestVector(1e14, 100, 'T'),
    TestVector(1e15, 1, 'P'),
    TestVector(1e16, 10, 'P'),
    TestVector(1e17, 100, 'P'),
    TestVector(1e18, 1, 'E'),
    TestVector(1e19, 10, 'E'),
    TestVector(1e20, 100, 'E'),
    TestVector(1e21, 1, 'Z'),
    TestVector(1e22, 10, 'Z'),
    TestVector(1e23, 100, 'Z'),
    TestVector(1e24, 1, 'Y'),
    TestVector(1e25, 10, 'Y'),
    TestVector(1e26, 100, 'Y'),
]  # test_vectors_metric_all_exponents


@pytest.mark.parametrize(
    "value, expected_scaled, expected_prefix",
    [(tv.value, tv.expected_scaled, tv.expected_prefix) for tv in test_vectors_binary_selected])
def test_convert_to_prefix_binary(value:float, expected_scaled: float, expected_prefix: str):
    scaled, prefix = convert_to_prefix(value, UnitPrefix.Binary)
    assert math.isclose(scaled, expected_scaled)
    assert prefix == expected_prefix


@pytest.mark.parametrize(
    "value, expected_scaled, expected_prefix",
    [(tv.value, tv.expected_scaled, tv.expected_prefix) for tv in test_vectors_metric_selected])
def test_convert_to_prefix_metric(value:float, expected_scaled: float, expected_prefix: str):
    scaled, prefix = convert_to_prefix(value, UnitPrefix.Metric)
    assert math.isclose(scaled, expected_scaled)
    assert prefix == expected_prefix


@pytest.mark.parametrize(
    "value, expected_scaled, expected_prefix",
    [(tv.value, tv.expected_scaled, tv.expected_prefix) for tv in test_vectors_binary_all_exponents])
def test_convert_to_prefix_all_exponents_binary(value:float, expected_scaled: float, expected_prefix: str):
    scaled, prefix = convert_to_prefix(value, UnitPrefix.Binary)
    assert math.isclose(scaled, expected_scaled)
    assert prefix == expected_prefix


@pytest.mark.parametrize(
    "value, expected_scaled, expected_prefix",
    [(tv.value, tv.expected_scaled, tv.expected_prefix) for tv in test_vectors_metric_all_exponents])
def test_convert_to_prefix_all_exponents_metric(value:float, expected_scaled: float, expected_prefix: str):
    scaled, prefix = convert_to_prefix(value, UnitPrefix.Metric)
    assert math.isclose(scaled, expected_scaled)
    assert prefix == expected_prefix


def test_convert_to_prefix_too_large_binary():
    value = test_vectors_binary_all_exponents[-1].value*1024
    with pytest.raises(ValueError):
        scaled, prefix = convert_to_prefix(value, UnitPrefix.Binary)


def test_convert_to_prefix_too_large_metric():
    value = test_vectors_metric_all_exponents[-1].value*1000
    with pytest.raises(ValueError):
        scaled, prefix = convert_to_prefix(value, UnitPrefix.Metric)


def test_convert_to_prefix_too_small_binary():
    with pytest.raises(ValueError):
        scaled, prefix = convert_to_prefix(0.1, UnitPrefix.Binary)


def test_convert_to_prefix_too_small_binary():
    value = test_vectors_metric_all_exponents[0].value/1000
    with pytest.raises(ValueError):
        scaled, prefix = convert_to_prefix(value, UnitPrefix.Binary)



#!/usr/bin/env python3
# vim: fileencoding=utf-8 ts=4
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: © Copyright 2025 by Christian Dönges. All rights reserved.
# SPDXID: SPDXRef-time-t-py
"""Timestamp conversion tool."""

import argparse
import enum
import sys

from pymisclib.enumaction import EnumAction
from pymisclib.utilities import convert_timestamp_to_datetime, initialize_console, iso8601_str


class TimestampFormat(enum.Enum):
    """Format of a timestamp."""
    Seconds = 's'
    Milliseconds = 'ms'
    Microseconds = 'µs'
    DotNet = '.net'


def parse_command_line_arguments(argv: list):
    """Parses the command line arguments specified in the argv array and
        returns an object containing all recognized options.

        Unrecognized options will be reported as an error.

        :param list argv: the command-line arguments as supplied by sys.argv[1:].
    """
    parser = argparse.ArgumentParser(description='Timestamp conversion tool.')
    parser.add_argument('timestamps', metavar='TIMESTAMPS',
                        type=str, nargs='+',
                        help='Timestamps to convert. Can be decimal (1, 3.4), or hexadecimal (0x345).')
    parser.add_argument('--format', '-f', action=EnumAction,
                        type=TimestampFormat, use='value',
                        default=TimestampFormat.Seconds,
                        help='Format of the timestamps.')
    args = parser.parse_args(argv)

    return args


def main(argv: list) -> int:
    """Do something useful."""
    # Initialize (Windows) console to use UTF-8.
    initialize_console()

    # Handle command line arguments.
    args = parse_command_line_arguments(argv[1:])

    # Do something useful here.
    if args.format == TimestampFormat.Seconds:
        seconds_digits = 0
    elif args.format == TimestampFormat.Milliseconds:
        seconds_digits = 3
    elif args.format == TimestampFormat.Microseconds:
        seconds_digits = 6
    elif args.format == TimestampFormat.DotNet:
        seconds_digits = 7
    else:
        raise ValueError(f'Timestamp format {args.format.name} is not supported.')

    nr_errors = 0
    timestamps = args.timestamps
    in_values = []
    max_len = 0
    for timestamp in timestamps:
        sd = seconds_digits
        try:
            if '.' in timestamp:
                ts = float(timestamp)
                sd = len(timestamp.split('.')[1])
            else:
                try:
                    ts = int(timestamp)
                except ValueError:
                    prefix = timestamp[:2].lower()
                    bases = {
                        '0b': 2,
                        '0o': 8,
                        '0x': 16
                    }
                    base = bases.get(prefix, 0)
                    if base < 2:
                        raise ValueError(f'unable to convert {timestamp} to a number')
                    ts = int(timestamp, base)
            in_values.append((ts, sd))
            max_len = max(max_len, len(str(ts)))
        except ValueError as e:
            print(f'ERROR: {timestamp} is not a number: {e}', file=sys.stderr)
            nr_errors += 1
            continue

    for ts, sd in in_values:
        try:
            dt = convert_timestamp_to_datetime(ts, args.format.value)
            print(f'{str(ts):{max_len}s} → {iso8601_str(dt, sd)}')
        except ValueError as e:
            nr_errors += 1
            print(f'ERROR: {ts} is not a valid timestamp: {e}', file=sys.stderr)

    return nr_errors


if __name__ == '__main__':
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# vim: fileencoding=utf8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2022, 2023, 2025 by Christian Dönges
# SPDXID: SPDXRef-littlejonny-py
"""Format output to a text table.

   In case you have been wondering, the name is a reference to
   https://xkcd.com/327/.


.. deprecated:: 1.6.0 Will be removed in 2.0.

"""

# Postpone type hint evaluation until after the entire class has been parsed.
# This was to become standard in 3.10 but will probably never be.
# @see https://stackoverflow.com/a/33533514
# @see https://mail.python.org/archives/list/python-dev@python.org/thread/CLVXXPQ2T2LQ5MP2Y53VVQFCXYWQJHKZ/
from __future__ import annotations


import enum
import logging
from dataclasses import dataclass, field
from typing import Any

from pymisclib.ansiterminal import AnsiControl, BgColor, FgColor, TextStyle
from pymisclib.unicodechars import UC


# Box drawing characters.
# The beauty of Namespace makes it possible to access a character like so:
# box_drawings.light.horizontal.
box_drawings = UC.box_drawings

# Type definition: a Text is a collection of lines, each of which is a string.
Text = list[str]
"""Type definition: a Text is a collection of lines, each of which is a string.

    The lines are organized by row, so the first string in the list is the first text line.
"""


@enum.unique
class LineStyle(enum.Enum):
    """Rendering style for lines."""
    double = box_drawings.double
    """Double line style ║."""
    light = box_drawings.light
    """Light single line style │."""
    heavy = box_drawings.heavy
    """Heavy single line style ┃."""


@enum.unique
class OutputFormat(enum.Enum):
    """Format for rendering the output."""
    ANSI = enum.auto()
    """Render for an ANSI terminal."""
    TEXT = enum.auto()
    """Render as plain text."""


@dataclass
class Point:
    """Coordinate in column (x) / row (y) space."""
    column: int = 1
    row: int = 1

    def __str__(self) -> str:
        """Return a string representation of the instance."""
        return f'({self.column};{self.row})'


@dataclass
class Box:
    """A box consisting of a frame with content.

        Constructor arguments:
            :param int origin_column: Top left corner column.
            :param int origin_row: Top left corner row.
            :param int width: Width in columns.
            :param int height: Height in rows.
            :param LineStyle line_style: Border line style.
            :param str border_style: Border color and font style expressed as ANSI format codes.
            :param str text_style: Content text color and font style expressed as ANSI format codes.
    """
    origin_column: int = 1  # starting x-coordinate
    """X-coordinate of the upper left corner of the box."""
    origin_row: int = 1  # starting y coordinate
    """Y-coordinate of the upper left corner of the box."""
    width: int = 40
    """Width of the box in columns."""
    height: int = 20
    """Height of the box in rows."""
    line_style: LineStyle = LineStyle.light
    """Style of the border characters."""
    border_style: str = FgColor.Black.value + BgColor.White.value
    """Rendering style (color and font style) of the border."""
    text_style: str = FgColor.Blue.value + BgColor.BrightYellow.value + TextStyle.Bold.value
    """Text style (color and font) of the content."""
    _text: Text = field(default_factory=list)

    @classmethod
    def make_from_points(cls, top_left: Point,
                         bottom_right: Point,
                         line_style: LineStyle = LineStyle.light) -> Box:
        """Create a new Box instance with the given top left and bottom right
        points.

        :param Point top_left: starting point in the upper left corner.
        :param Point bottom_right: ending point (inclusive) in the lower right
            corner.
        :param LineStyle line_style: Style to use when rendering the box.
        :return: An initialized instance.
        :rtype Box:
        """
        if top_left.column < 1:
            raise ValueError('left column too small')
        if top_left.row < 1:
            raise ValueError('top row too small')
        if bottom_right.column <= top_left.column:
            raise ValueError('right column too small')
        if bottom_right.row <= top_left.row:
            raise ValueError('bottom column too small')
        width = bottom_right.column - top_left.column
        height = bottom_right.row - top_left.row
        return Box(top_left.row, top_left.column, width, height, line_style)

    def flow_text(self, text: str, truncate: bool = True):
        """Flow the given text in to the current box.

        :param str text: Text to flow into the box.
        :param bool truncate: If False, raise an exception if text won't fit into
            the box. If True, truncate the text.
        """
        self.text = flow_text(text, self.width, self.height, truncate)


    @property
    def text(self) -> Text:
        """Text contained in the box."""
        return self._text

    @text.setter
    def text(self, text: Text|str, truncate: bool = True):
        """Set the text contained in the box.

        If the text is a Text, it will simply be fit into the box. If it is a
        string, it will be flowed into the box.

        :param str|Text text: Text to flow into the box.
        :param bool truncate: If False, raise an exception if text won't fit.
        :raise ValueError: Raised if the text won't fit into the box and
            truncate is set to False.
        """
        self._text = []
        if isinstance(text, str):
            text = flow_text(text, self.width, self.height, truncate)
        if len(text) > self.height:
            if truncate:
                edited_text = text[:self.height]
            else:
                raise ValueError('too many lines')
        else:
            edited_text = text
        for line in edited_text:
            if len(line) > self.width:
                if truncate:
                    line = line[:self.width]
                else:
                    ValueError('line too wide')
            else:
                line = line + ' ' * (self.width - len(line))
            self._text.append(line)

    def render(self, output_format: OutputFormat) -> Text:
        """Render the box in the given output_format.

        :param OutputFormat output_format: Rendering format.
        :return: A list of lines rendering the instance.
        :rtype Text:
        """
        if output_format == OutputFormat.ANSI:
            return self._render_ansi()
        elif output_format == OutputFormat.TEXT:
            return self._render_text()
        raise ValueError('unknown output output_format')

    def _render_ansi(self) -> Text:
        """Render the instance on an ANSI terminal."""
        lines = []
        if self.origin_column > 1:
            prefix = AnsiControl.move_cursor_forward(self.origin_column - 1)
        else:
            prefix = ''
        box = self.line_style.value
        s = self.border_style + box.down_and_right + box.horizontal * self.width + \
            box.down_and_left + TextStyle.Reset.value
        if self.origin_column > 1 or self.origin_row > 1:
            s = AnsiControl.move_cursor_to(self.origin_column, self.origin_row) + s
        else:
            s = prefix + s
        lines.append(s)
        s = prefix + self.border_style + box.vertical + \
            self.text_style + (' ' * self.width) + TextStyle.Reset.value + \
            self.border_style + box.vertical + TextStyle.Reset.value
        for i in range(self.height):
            if i < len(self._text):
                lines.append(prefix + self.border_style + box.vertical +
                             self.text_style + self._text[i] + TextStyle.Reset.value +
                             self.border_style + box.vertical + TextStyle.Reset.value)
            else:
                lines.append(s)
        s = prefix + self.border_style + box.up_and_right + \
            box.horizontal * self.width + box.up_and_left + TextStyle.Reset.value
        lines.append(s)

        return lines

    def _render_text(self) -> Text:
        """Render the instance on dumb terminal (e.g. as plain text).

        The origin column and row are created using spaces and blank lines.

        :return: A list of lines, each of which is a string that contains no
            formatting characters.
        :rtype list[str]:
        """
        lines = []
        if self.origin_row > 1:
            for i in range(self.origin_row - 1):
                lines.append('')
        prefix = ' ' * (self.origin_column - 1)
        box = self.line_style.value
        s = prefix + box.down_and_right + box.horizontal * self.width + box.down_and_left
        lines.append(s)
        s = prefix + box.vertical + ' ' * self.width + box.vertical
        for i in range(self.height):
            if i < len(self._text):
                lines.append(prefix + box.vertical + self._text[i] + box.vertical)
            else:
                lines.append(s)
        s = prefix + box.up_and_right + box.horizontal * self.width + box.up_and_left
        lines.append(s)

        return lines


@dataclass
class ColumnTable:
    """Table with data arranged by column.

    .. deprecated:: 2.0.0
    """
    _headings: list[str] = field(default_factory=list)  # row or column heading
    _formats: list[str] = field(default_factory=list)  # cell formats
    _cells: list[list[Any]] = field(default_factory=list)  # column[row]
    _num_columns: int = 0
    _num_rows: int = 0

    @property
    def cell_formats(self) -> list[str]:
        """Return cell formats by columns."""
        return self._formats

    @property
    def headings(self) -> list[str]:
        """Return headings by column."""
        return self._headings

    @property
    def num_columns(self):
        """Number of columns in the table."""
        return self._num_columns

    @property
    def num_rows(self):
        """Number of rows in the table."""
        return self._num_rows

    def set_table(self,
                  headings: list[str],
                  cell_formats: list[str],
                  cells: list[list[str]]):
        """Set the content of the table with cells specified by column.

            :param list[str] headings: List of headings.
            :param list[str] cell_formats: List of cell formats for each column.
                The formats use the same format-language as Python format strings.
            :param list[str] cells: Matrix of cell contents, specified column[row].

            Example
            ---------

            .. code-block:: python

                ct2 = ColumnTable()
                ct2.set_table(
                    headings=['First', 'Second', 'Third', 'Fourth'],
                    cell_formats=['3d', '08x', 's', '>s'],
                    cells=[[1, 2, 12, 123, 1000],
                           [0x12345678, 0xffffee01, -1, 0, 23],
                           ['abc def geh', '12345', 'minus one', 'zero', 'First'],
                           ['this is a sample text', 'Short', 'negative hex number',
                            'Zero hexadecimal number', 'The first column is too large.']]
                )
                print_lines(ct2.draw(LineStyle.light))

            Output::

                ┌───────┬──────────┬─────────────┬────────────────────────────────┐
                │ First │ Second   │ Third       │ Fourth                         │
                ├───────┼──────────┼─────────────┼────────────────────────────────┤
                │   1   │ 12345678 │ abc def geh │ this is a sample text          │
                │   2   │ ffffee01 │ 12345       │ Short                          │
                │  12   │ -0000001 │ minus one   │ negative hex number            │
                │ 123   │ 00000000 │ zero        │ Zero hexadecimal number        │
                │ 1000  │ 00000017 │ First       │ The first column is too large. │
                └───────┴──────────┴─────────────┴────────────────────────────────┘
        """
        if len(headings) != len(cell_formats):
            raise ValueError('number columns in headings and formats do not match')
        if len(headings) != len(cells):
            raise ValueError('number of columns in headings and cells do not match')
        self._num_columns = len(headings)
        self._num_rows = len(cells[0])
        self._headings = headings
        self._formats = cell_formats
        self._cells = cells

    def set_table_transposed(self,
                             headings: list[str],
                             cell_formats: list[str],
                             cells: list[list[str]]):
        """Set the content of the table with cells specified by row.

            :param list[str] headings: List of headings.
            :param list[str] cell_formats: List of cell formats for each column.
            :param list[str] cells: Matrix of cell contents, specified row[column].

            Example
            ---------

            .. code-block:: python

                ct1 = ColumnTable()
                ct1.set_table_transposed(
                    headings=['First', 'Second', 'Third', 'Fourth'],
                    cell_formats=['3d', '08x', 's', '>s'],
                    cells=[[1, 0x12345678, 'abc def geh', 'this is a sample text'],
                           [2, 0xffffee01, '12345', 'Short'],
                           [12, -1, 'minus one', 'negative hex number'],
                           [123, 0, 'zero', 'Zero hexadecimal number'],
                           [1000, 23, 'First', 'The first column is too large.']]
                )
                print_lines(ct1.draw(LineStyle.light))

            Output::

                ┌───────┬──────────┬─────────────┬────────────────────────────────┐
                │ First │ Second   │ Third       │ Fourth                         │
                ├───────┼──────────┼─────────────┼────────────────────────────────┤
                │   1   │ 12345678 │ abc def geh │ this is a sample text          │
                │   2   │ ffffee01 │ 12345       │ Short                          │
                │  12   │ -0000001 │ minus one   │ negative hex number            │
                │ 123   │ 00000000 │ zero        │ Zero hexadecimal number        │
                │ 1000  │ 00000017 │ First       │ The first column is too large. │
                └───────┴──────────┴─────────────┴────────────────────────────────┘
        """
        transposed_cells = [list(x) for x in zip(*cells)]
        self.set_table(headings, cell_formats, transposed_cells)

    def draw(self, style: LineStyle) -> Text:
        """Render the table to a list of lines.

            :param LineStyle style: The line style to use for the table.
            :rtype Text:
            :return: A list of rows which make up the rendered table.
        """
        formatted_cells = []
        column_widths = []
        for c in range(self.num_columns):
            column_width = len(self._headings[c])
            column_data = self._cells[c]
            fmt = self._formats[c]
            formatted_column = []
            for cell in column_data:
                formatted_cell = f'{cell:{fmt}}'
                formatted_column.append(formatted_cell)
                column_width = max(column_width,  len(formatted_cell))
            formatted_cells.append(formatted_column)
            column_widths.append(column_width)

        box = style.value
        lines = []
        first = True
        s = ''
        for c in range(self.num_columns):
            column_width = column_widths[c]
            if first:
                s = box.down_and_right + box.horizontal * (column_width + 2)
                first = False
            else:
                s += box.down_and_horizontal + box.horizontal * (column_width + 2)
        s += box.down_and_left
        lines.append(s)

        s = ''
        for c in range(self.num_columns):
            column_width = column_widths[c]
            s += box.vertical + f' {self._headings[c]:{column_width}s} '
        s += box.vertical
        lines.append(s)

        first = True
        s = ''
        for c in range(self.num_columns):
            column_width = column_widths[c]
            if first:
                s = box.vertical_and_right + box.horizontal * (column_width + 2)
                first = False
            else:
                s += box.vertical_and_horizontal + box.horizontal * (column_width + 2)
        s += box.vertical_and_left
        lines.append(s)

        for r in range(self.num_rows):
            s = ''
            for c in range(self.num_columns):
                column_width = column_widths[c]
                s += box.vertical + f' {formatted_cells[c][r]:{column_width}s} '
            s += box.vertical
            lines.append(s)

        first = True
        s = ''
        for c in range(self.num_columns):
            column_width = column_widths[c]
            if first:
                s = box.up_and_right + box.horizontal * (column_width + 2)
                first = False
            else:
                s += box.up_and_horizontal + box.horizontal * (column_width + 2)
        s += box.up_and_left
        lines.append(s)

        return lines


@enum.unique
class Alignment(enum.Enum):
    """Content alignment within a cell."""
    Center = '^'
    """Align content in the center of the cell."""
    Left = '<'
    """Align content to the left of the cell."""
    Right = '>'
    """Align content to the right of the cell."""


@dataclass
class CellStyle:
    """Styling to use for the cell."""
    heading_style: str = TextStyle.Bold.value
    """Heading cell content style as a string of ANSI escape codes.
    Default is **bold**.
    """
    plain_style: str = ''
    """Plain cell content style as a string of ANSI escape codes.
    Default is plain.
    """


@dataclass
class TableCell:
    """A single cell of a table.

    Constructor arguments:
        :param Any content: Content of the cell. Usually a string or number.
        :param str fmt: f-string `format_spec` to format the content.
        :param bool is_heading: True if the cell is a heading cell.
        :param Alignment alignment: Alignment of the formatted content in the
            cell.
        :param CellStyle style: Style of the rendered content.
    """
    content: Any = ''  # Content that is rendered.
    fmt: str = 's'  # Format string to render the content.
    is_heading: bool = False  # True to render as heading.
    alignment: Alignment = Alignment.Right
    style: CellStyle = None
    join_right: bool = False  # Join with the cell to the right.
    join_bottom: bool = False  # Join with the cell below.

    def render(self,
               output_format: OutputFormat = OutputFormat.TEXT,
               width: int = 0) -> Text:
        """Render the box in the given output_format.

        :param OutputFormat output_format: Rendering format.
        :param int width: Width of the cell. If zero, the width is unconstrained
            and the content will not be aligned.
        :return: A list of lines representing the instance.
        :rtype: Text
        """
        if output_format == OutputFormat.ANSI:
            return self._render_ansi(width)
        elif output_format == OutputFormat.TEXT:
            return self._render_text(width)
        raise ValueError('unknown output output_format')

    def _render_ansi(self, width: int) -> str:
        """Render the content as a text string with ANSI formatting.

        :param int width: Width of the cell.
        :return: ANSI representation of the cell.
        :rtype: str
        """
        s = self._render_text(width)
        if self.style is not None:
            if self.is_heading:
                s = self.style.heading_style + s + TextStyle.Reset.value
            else:
                s = self.style.plain_style + s + TextStyle.Reset.value
        return s

    def _render_text(self, width: int) -> str:
        """Render the content as a text string.

        :param int width: Width of the cell.
        :return: Text representation of the cell.
        :rtype: str
        """
        try:
            formatted_str = f'{self.content:{self.fmt}}'
        except ValueError:
            formatted_str = str(self.content)

        if width:
            aligned_str = f'{formatted_str:{self.alignment.value}{width}s}'[:width]
        else:
            aligned_str = formatted_str
        return aligned_str

    def set(self, value: Any,
            fmt: str = '%s',
            justify: Alignment = Alignment.Right,
            is_heading: bool = False):
        """Set the value of the content and the format string for rendering.

        :param Any value: Content of the cell.
        :param str fmt: Format string for the cell value. Uses f-string syntax.
        :param Alignment justify: Cell justification after formatting.
        :param bool is_heading: True if cell gets heading formatting, False otherwise.
        """
        self.content = value
        self.fmt = fmt
        self.alignment = justify
        self.is_heading = is_heading


@dataclass
class Table:
    """Table made up of a rectangle of :py:class:`TableCell` instances.

    Constructor arguments:
        :param list[list[TableCell]] | None cells: Two-dimensional list of
            :py:class:`TableCell` instances making up the table. The cells are
            organized by row and column (.e.g. `cells[row][column]`).
        :param logging.Logger logger: Logger instance for diagnostics.
        :param CellsStyle default_style: All new cells start with this style.
    """
    cells: list[list[TableCell]] | None = None
    """Cells, ordered by row and column, e.g. cells[row][column]."""
    logger: logging.Logger = logging.getLogger(__name__)
    """Logger instance for diagnostics."""
    default_style: CellStyle = field(default_factory=lambda: CellStyle())
    """Style all cells using this style unless specified otherwise."""
    column_widths: list[int] = field(default_factory=list)  # 0 to auto-size

    def __post_init__(self):
        """Initializations performed after the constructor was called."""
        if len(self.column_widths) == 0:
            self.column_widths = [0] * self.width
        if self.width != len(self.column_widths):
            raise ValueError(f'Table has {self.width} columns, '
                             f'{len(self.column_widths)} column widths given.')

    @property
    def height(self) -> int:
        """Number of rows in the table."""
        if self.cells is None:
            return 0
        return len(self.cells)

    @property
    def width(self) -> int:
        """Number of columns in the table."""
        if self.cells is None:
            return 0
        return len(self.cells[0])

    def cell_at(self, column: int, row: int) -> TableCell:
        """Return the cell at the given location.

        :param int column: Column number of the cell.
        :param int row: Row number of the cell.
        :return: Cell of the table at the given location.
        :rtype: TableCell
        :raises IndexError: Row or column was invalid.
        """
        if column < 0 or column >= self.width:
            raise IndexError('invalid column number')
        if row < 0 or row >= self.height:
            raise IndexError('invalid row number')
        return self.cells[row][column]

    def column(self, column_idx: int) -> list[TableCell]:
        """Return the specified column of the table.

        :param int column_idx: Column index [0..table.width[.
        :return: List of cells in the column ordered by row.
        :rtype list[TableCell]:
        :raises IndexError: Column was invalid
        """
        if column_idx < 0 or column_idx >= self.height:
            raise IndexError(f'invalid column number {column_idx}')
        column = [r[column_idx] for r in self.cells]
        return column

    def column_width(self, column_idx: int) -> int:
        """Return the configured width of the specified column.

        :param int column_idx: Column index [0..table.width[.
        :return: Configured width of the column, or 0 to auto-size.
        :rtype: int
        :raises IndexError: Column was invalid
        """
        if column_idx < 0 or column_idx >= self.height:
            raise IndexError(f'invalid column number {column_idx}')
        return self.column_widths[column_idx]

    def populate(self, content: list[list[Any]],
                 column_formats: list[str],
                 column_alignment: list[str] | None = None):
        """Resize and populate a table with the given content.

        """
        self.logger.debug('%s.populate(%d, %d)', len(content), len(column_formats))
        # Check arguments for consistency.
        num_columns = len(column_formats)
        if num_columns == 0:
            raise IndexError('no columns')
        if len(content) == 0:
            raise IndexError('no rows')
        for row in content:
            if len(row) != num_columns:
                self.logger.error('Expected %d columns, found %d.', num_columns, len(row))
                raise IndexError('content wrong number of columns')
        if column_alignment is None:
            column_alignment = [Alignment.Left] * num_columns
        elif len(column_alignment) != num_columns:
            self.logger.error('Expected %d columns, found column_alignment:%d.', num_columns, len(column_alignment))
            raise IndexError('column_alignment wrong number of columns')
        self.resize(num_columns, len(content))
        for row_idx, row in enumerate(content):
            for column_idx, cell_content in enumerate(row):
                self.cells[row_idx][column_idx].set(cell_content,
                                                    column_formats[column_idx],
                                                    column_alignment[column_idx])
        self.column_widths = [0] * num_columns

    def render(self,
               output_format: OutputFormat = OutputFormat.TEXT,
               line_style: LineStyle = LineStyle.light) -> Text:
        """Render the table in the given format to a list of strings.

        :param OutputFormat output_format: Format to render to.
        :param LineStyle line_style: Style used to render cell borders.
        :return: Rendered repsentation of the instance.
        :rtype: Text
        """
        self.logger.debug('%s.render(%s, %s)', self.__class__.__name__,
                          output_format.name, line_style.name)
        lines = []
        column_widths = [0] * self.width  # width in characters of each column
        for column_idx in range(self.width):
            tcw = self.column_widths[column_idx]
            if tcw > 0:
                # Use configured width.
                column_widths[column_idx] = tcw
            else:
                # Use maximum width of any cell in the column.
                col = self.column(column_idx)
                for cell in col:
                    rc = cell.render(output_format=OutputFormat.TEXT)
                    column_widths[column_idx] = max(column_widths[column_idx], len(rc))

        bs = line_style.value
        previous_row = None
        for row_idx in range(self.height):
            row = self.cells[row_idx]
            # Render the top cell border.
            s = ''
            joining_right = False
            joining_previous_right = False
            for column_idx in range(self.width):
                column_width = column_widths[column_idx]
                if row_idx == 0:
                    if column_idx == 0:
                        s = bs.down_and_right + bs.horizontal * (column_width + 2)
                    else:
                        if joining_right:
                            s += bs.horizontal * (column_width + 3)
                        else:
                            s += bs.down_and_horizontal + bs.horizontal * (column_width + 2)
                else:
                    if column_idx == 0:
                        s = bs.vertical_and_right + bs.horizontal * (column_width + 2)
                    else:
                        if joining_right:
                            if joining_previous_right:
                                s += bs.horizontal * (column_width + 3)
                            else:
                                s += bs.up_and_horizontal + bs.horizontal * (column_width + 2)
                        else:
                            if joining_previous_right:
                                s += bs.down_and_horizontal + bs.horizontal * (column_width + 2)
                            else:
                                s += bs.vertical_and_horizontal + bs.horizontal * (column_width + 2)
                cell = row[column_idx]
                joining_right = cell.join_right
                if previous_row:
                    joining_previous_right = previous_row[column_idx].join_right
            if row_idx == 0:
                s+= bs.down_and_left
            else:
                s += bs.vertical_and_left
            lines.append(s)

            # Render the cell side borders and content.
            s = ''
            joining_right = False
            for column_idx in range(self.width):
                cell = row[column_idx]
                if joining_right:
                    # This is the nth cell of the join.
                    column_width += column_widths[column_idx] + 3
                else:
                    column_width = column_widths[column_idx]
                    render_cell = cell
                if cell.join_right:
                    # Defer rendering until we reach the end of the joined section.
                    joining_right = True
                    continue
                rc = render_cell.render(output_format, width=column_width)
                s += bs.vertical + f' {rc} '
                joining_right = False
            s += bs.vertical
            lines.append(s)
            previous_row = row

        # Render the bottom table border.
        s = ''
        joining_right = False
        for column_idx in range(self.width):
            column_width = column_widths[column_idx]
            if column_idx == 0:
                s = bs.up_and_right + bs.horizontal * (column_width + 2)
            else:
                if joining_right:
                    s += bs.horizontal * (column_width + 3)
                else:
                    s += bs.up_and_horizontal + bs.horizontal * (column_width + 2)
            cell = row[column_idx]
            joining_right = cell.join_right
        s += bs.up_and_left
        lines.append(s)
        return lines


    def resize(self, width: int, height: int):
        """Resize the table to the given width and height.

        :param int width: Number of columns in the table.
        :param int height: Number of rows in the table.
        """
        self.logger.debug('%s.resize(%d, %d)', self.__class__.__name__, width, height)

        # Error checking.
        if width < 0:
            raise ValueError('width must be non-negative')
        if height < 0:
            raise ValueError('height must be non-negative')
        if width == 0 or height == 0:
            if width != height:
                raise ValueError('width and height must both be zero')
            self.cells = None
            return

        if height < self.height:
            self.logger.debug('Reduce number of rows from %d to %d.', self.height, height)
            self.cells = self.cells[:height]
        else:
            num_rows_to_add = height - self.height
            if num_rows_to_add > 0:
                self.logger.debug('Extend number of rows from %d to %d.', self.height, height)
                num_columns = self.width
                if self.cells is None:
                    self.cells = []
                for i in range(num_rows_to_add):
                    row = []
                    for j in range(num_columns):
                        row.append(TableCell(style=self.default_style))
                    self.cells.append(row)
        if width < self.width:
            self.logger.debug('Reduce number of columns from %d to %d.', self.width, width)
            for r in range(self.height):
                self.cells[r] = self.cells[r][:width]
            self.column_widths = self.column_widths[:width]
        else:
            num_columns_to_add = width - self.width
            if num_columns_to_add > 0:
                self.logger.debug('Extend number of columns from %d to %d.', self.width, width)
                for r in range(self.height):
                    for c in range(num_columns_to_add):
                        self.cells[r].append(TableCell(style=self.default_style))
                self.column_widths += [0] * num_columns_to_add

    def row(self, row_idx: int) -> list[TableCell]:
        """Return the specified row of the table.

        :param int row_idx: Row index [0..table.height[.
        :return: List of cells in the row, ordered by column.
        :rtype list[TableCell]:
        :raises IndexError: Column was invalid
        """
        if row_idx < 0 or row_idx >= self.height:
            raise IndexError(f'invalid row number {row_idx}')
        row = self.cells[row_idx]
        return row

    def set_column_width(self, column_idx: int, width: int):
        """Set the width in characters of the given column..

        :param int column_idx: Column index [0..table.width[.
        :param int width: Width in characters or 0 to auto-size.
        :raises IndexError: Column was invalid.
        :raises ValueError: Width was less than zero.
        """
        if column_idx < 0 or column_idx >= self.height:
            raise IndexError(f'invalid column number {column_idx}')
        if width < 0:
            raise ValueError(f'invalid width {width}')
        self.column_widths[column_idx] = width


def draw_box(x, y, style) -> Text:
    """Construct a box of size x * y with style."""
    box = style.value

    lines = []
    s = box.down_and_right + box.horizontal * x + box.down_and_left
    lines.append(s)
    s = box.vertical + ' ' * x + box.vertical
    for i in range(y):
        lines.append(s)
    s = box.up_and_right + box.horizontal * x + box.up_and_left
    lines.append(s)

    return lines


def flow_text(text: str,
              width: int = 0,
              height: int = 0,
              truncate: bool = True) -> Text:
    """Flow the given text in to the given space.

    :param str text: Text to flow into the box.
    :param int width: Number of characters in a line to flow text into.
    :param int height: Number of lines to flow text into.
    :param bool truncate: If False, raise an exception if text won't fit into
        the box. If True, truncate the text.
    :rtype: Text
    :return: Flowed text, one string per line, from top to bottom.
    """
    if width < 0:
        raise ValueError('width must be positive or zero')
    if height < 0:
        raise ValueError('height must be positive or zero')
    if width == 0:
        # No width constraint means we can go as wide as we want.
        return [text]
    lines: list[str] = []
    start_offset = 0
    max_offset = len(text) - 1
    while start_offset < max_offset:
        end_of_line_offset = min(start_offset + width, max_offset)
        skip_space = 1
        if start_offset + width < max_offset:
            end_of_line_offset = start_offset + width
            wrap_offset = text.rfind(' ', start_offset, end_of_line_offset)
        else:
            end_of_line_offset = max_offset
            wrap_offset = end_of_line_offset + 1
        if wrap_offset < 0:
            # No space found to wrap line.
            if truncate:
                wrap_offset = end_of_line_offset
                skip_space = 0
            else:
                raise ValueError('unable to wrap text')
        line = text[start_offset:wrap_offset]
        lines.append(line)
        start_offset = wrap_offset + skip_space
    if height != 0 and len(lines) > height:
        if truncate:
            lines = lines[:height]
        else:
            ValueError(f'too many lines of text: {len(lines)} > {height}')
    return lines


def print_lines(lines: Text):
    """Print lines stored as a list of strings to stdout."""
    for line in lines:
        print(line)


def demo_table_simple():
    """Demonstrate the use of the Table class."""
    print('Demonstrate simple use of the Table class.')
    cell_alignment = [Alignment.Left,
                    Alignment.Center,
                    Alignment.Right,
                    Alignment.Center,
                    Alignment.Left]
    cell_formats=['3d', '08x', 's', '>s', '<s']
    cell_contents=[
       [j.name for j in cell_alignment],
       [1, 2, 12, 123, 1000],
       [0x12345678, 0xffffee01, -1, 0, 23],
       ['abc def geh', 12345, 'minus one', 'zero', 'First'],
       ['this is a sample text', -5, 'negative hex number',
        'Zero hexadecimal number', 'The first column is too large.'],
       [97.2, 123, '***', '***', '***'],]

    for i in range(3):
        print()
    style1 = CellStyle(heading_style=TextStyle.Reverse.value, plain_style='')
    style2 = CellStyle(
        heading_style=TextStyle.Bold.value + BgColor.Red.value + FgColor.White.value,
        plain_style=TextStyle.Italic.value + BgColor.BrightBlue.value)
    t = Table(default_style=style1)
    t.populate(content=cell_contents,
               column_formats=cell_formats,
               column_alignment=cell_alignment)
    # Override header formatting.
    for col_idx in range(t.width):
        c = t.cell_at(column=col_idx, row=0)
        c.fmt = 's'
        c.alignment = Alignment.Center
        c.is_heading = True
    lines = t.render(OutputFormat.TEXT, LineStyle.light)
    print_lines(lines)
    lines = t.render(OutputFormat.ANSI, LineStyle.double)
    print_lines(lines)
    # Override styling of diagonal cells.
    for i in range(min(t.height, t.width)):
        t.cell_at(i, i).style = style2
    lines = t.render(OutputFormat.ANSI, LineStyle.heavy)
    print_lines(lines)


def demo_table_advanced():
    """Demonstrate advanced use of the Table class."""
    print('Demonstrate advanced use of the Table class.')
    ds = CellStyle(TextStyle.Bold.value + TextStyle.Italic.value, '')
    cells = [ [
        TableCell('First', is_heading=True, alignment=Alignment.Left, style=ds),
        TableCell('Second', is_heading=True, alignment=Alignment.Center,
                  style=ds, join_right=True),
        TableCell('Third', is_heading=True, alignment=Alignment.Left, style=ds),
        TableCell('Fourth', is_heading=True, alignment=Alignment.Right, style=ds),
        TableCell('Fifth', is_heading=True, alignment=Alignment.Left, style=ds),
    ] ]  # cells
    cells.append([
        TableCell(1, '3d', alignment=Alignment.Right, style=ds),
        TableCell('This is a text that will not fit in the table.'
                  ' This is because it is too long for a single cell.',
                  alignment=Alignment.Center, style=ds, join_right=True),
        TableCell('3', join_right=True),
        TableCell(4, '5d'),
        TableCell('***', alignment=Alignment.Left, style=ds),
    ])
    cells.append([
        TableCell(2, '3d', alignment=Alignment.Right, style=ds),
        TableCell('This will fit easily.',
                  alignment=Alignment.Center, style=ds),
        TableCell('First time'),
        TableCell(44, '5d'),
        TableCell('***', alignment=Alignment.Center, style=ds),
    ])
    cells.append([
        TableCell(3, '3d', alignment=Alignment.Right, style=ds, join_right=True),
        TableCell('Not seen', style=ds),
        TableCell('2nd time', alignment=Alignment.Left, style=ds),
        TableCell(444, '5d', style=ds),
        TableCell('***', alignment=Alignment.Right, style=ds),
    ])
    cells.append([
        TableCell(4, '3d', alignment=Alignment.Right, style=ds, join_right=True),
        TableCell('not seen', style=ds, join_right=True),
        TableCell('not seen', style=ds),
        TableCell(4444, '5d', style=ds),
        TableCell('*', alignment=Alignment.Left, style=ds),
    ])
    cells.append([
        TableCell(5, '3d', alignment=Alignment.Right, style=ds, join_right=True),
        TableCell('not seen', style=ds, join_right=True),
        TableCell('not seen', style=ds, join_right=True),
        TableCell(44444, '5d', style=ds),
        TableCell(' *', alignment=Alignment.Left, style=ds),
    ])
    cells.append([
        TableCell(6, '3d', alignment=Alignment.Right, style=ds, join_right=True),
        TableCell('not seen', style=ds, join_right=True),
        TableCell('not seen', style=ds, join_right=True),
        TableCell(444444, '5d', style=ds, join_right=True),
        TableCell('  *', alignment=Alignment.Left, style=ds),
    ])
    t = Table(cells, default_style=ds)
    t.set_column_width(1, 10)
    lines = t.render(OutputFormat.ANSI, LineStyle.light)
    print_lines(lines)
    lines = t.render(OutputFormat.TEXT, LineStyle.heavy)
    print_lines(lines)


def demo():
    """Littlejonny demo."""
    print_lines(draw_box(5, 3, LineStyle.light))
    print_lines(draw_box(1, 1, LineStyle.double))
    print_lines(draw_box(40, 3, LineStyle.heavy))

    b1 = Box(8, 3, 17, 9, LineStyle.light,
             FgColor.Green.value + BgColor.White.value,
             FgColor.Blue.value + BgColor.BrightYellow.value + TextStyle.Italic.value)
    b1.text = [
        '(8;7)      (17;8)',
        '',
        '',
        '',
        '        *       ',
        '',
        '',
        '',
        '(8;16)    (17;16)',
    ]
    b2 = Box(1, 1, 64, 13, LineStyle.double,
             FgColor.Black.value + BgColor.Green.value,
             FgColor.Red.value + BgColor.BrightWhite.value + TextStyle.Bold.value)
    b2.text = [
        '         1         2         3         4         5         6    ',
        '1234567890123456789012345678901234567890123456789012345678901234',
        '',
        '                                                           right',
        'left',
        '                               center',
        UC.box_drawings.light.horizontal * b2.width,
        'ABC DEF GHI JKL MNO PRQ STU VWX YZ!',
        'ABC DEF GHI JKL MNO PRQ STU VWX YZ!',
        UC.box_drawings.light.horizontal * b2.width,
    ]
    print_lines(b1.render(OutputFormat.TEXT))
    print_lines(b2.render(OutputFormat.TEXT))

    b3 = Box(4, 4, 20, 9)
    b3.flow_text('This is a really long text to flow into the box that is'
                 ' just a tiny little thing of a box. '
                 'OMG_this_will_never_fit_into_a_single_line'
                 ' :)')
    print_lines(b3.render(OutputFormat.TEXT))

    # row[column]
    ct1 = ColumnTable()
    ct1.set_table_transposed(
        headings=['First', 'Second', 'Third', 'Fourth'],
        cell_formats=['3d', '08x', 's', '>s'],
        cells=[[1, 0x12345678, 'abc def geh', 'this is a sample text'],
               [2, 0xffffee01, '12345', 'Short'],
               [12, -1, 'minus one', 'negative hex number'],
               [123, 0, 'zero', 'Zero hexadecimal number'],
               [1000, 23, 'First', 'The first column is too large.']]
    )
    print_lines(ct1.draw(LineStyle.light))

    # column[row]
    ct2 = ColumnTable()
    ct2.set_table(
        headings=['First', 'Second', 'Third', 'Fourth'],
        cell_formats=['3d', '08x', 's', '>s'],
        cells=[[1, 2, 12, 123, 1000],
               [0x12345678, 0xffffee01, -1, 0, 23],
               ['abc def geh', '12345', 'minus one', 'zero', 'First'],
               ['this is a sample text', 'Short', 'negative hex number',
                'Zero hexadecimal number', 'The first column is too large.']]
    )
    print_lines(ct2.draw(LineStyle.light))
    print_lines(b1.render(OutputFormat.ANSI))
    print_lines(b2.render(OutputFormat.ANSI))

    demo_table_simple()
    demo_table_advanced()


if __name__ == '__main__':
    demo()



#!/usr/bin/env python3
# vim: fileencoding=utf8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2023 by Christian Dönges
# SPDXID: SPDXRef-init-py

__title__ = "pymisclib"
__description__ = "Collection of helpful stuff."
__url__ = "https://rot.regenbogenkraut.de/cd/pymisclib"
__author__ = "Christian Dönges"
__author_email__ = "cd.pymisclib@platypus-projects.net"
__license__ = "Apache 2.0"
__copyright__ = "© Copyright 2012-2024 by Christian Dönges <cd.pymisclib@platypus-projects.net>"

from .__version__ import (
    __version__,
    __version_tuple__,
)



#!/usr/bin/env python3
# vim: fileencoding=utf-8 ts=4
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: © Copyright 2024 by Christian Dönges. All rights reserved.
# SPDXID: SPDXRef-test-dolcenum-py

import enum

import pytest
from pymisclib.dolcenum import DolcEnumMeta, MissingDocstringError


def test_no_documentation():
    """No documentation at all is allowed."""
    class E(enum.IntFlag, metaclass=DolcEnumMeta):
        b0 = 1
        b1 = 2
        b2 = 4


def test_all_docstrings():
    """Class and all members have a docstring is a good case."""
    class E(enum.IntFlag, metaclass=DolcEnumMeta):
        """Class E docstring."""
        b0 = 1
        """Member E.b0 docstring."""
        b1 = 2
        """Member E.b1 docstring."""
        b2 = 4
        """Member E.b2 docstring."""

    assert E.__doc__ == 'Class E docstring.'
    assert E.b0.__doc__ == 'Member E.b0 docstring.'
    assert E.b1.__doc__ == 'Member E.b1 docstring.'
    assert E.b2.__doc__ == 'Member E.b2 docstring.'


def test_class_docstring_only():
    with pytest.raises(MissingDocstringError):
        class E(enum.IntFlag, metaclass=DolcEnumMeta):
            """Class docstring."""
            b0 = 1
            b1 = 2
            b2 = 4


def test_member_docstring_only():
    """Missing class docstring is a bad case."""
    with pytest.raises(MissingDocstringError):
        class E(enum.IntFlag, metaclass=DolcEnumMeta):
            b0 = 1
            """Member E.b0 docstring."""
            b1 = 2
            """Member E.b1 docstring."""
            b2 = 4
            """Member E.b2 docstring."""


def test_missing_one_docstring():
    with pytest.raises(MissingDocstringError):
        class E(enum.IntFlag, metaclass=DolcEnumMeta):
            """Class E docstring."""
            b0 = 1
            b1 = 2
            """Member E.b1 docstring."""
            b2 = 4
            """Member E.b2 docstring."""

    with pytest.raises(MissingDocstringError):
        class E(enum.IntFlag, metaclass=DolcEnumMeta):
            """Class E docstring."""
            b0 = 1
            """Member E.b0 docstring."""
            b1 = 2
            b2 = 4
            """Member E.b2 docstring."""

    with pytest.raises(MissingDocstringError):
        class E(enum.IntFlag, metaclass=DolcEnumMeta):
            """Class E docstring."""
            b0 = 1
            """Member E.b0 docstring."""
            b1 = 2
            """Member E.b1 docstring."""
            b2 = 4



#!/usr/bin/env python3
# vim: fileencoding=utf-8 ts=4
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: © Copyright 2024, 2025 by Christian Dönges. All rights reserved.
# SPDXID: SPDXRef-hardwareregister-py
"""Description of hardware registers.

    A register has a name and an offset (which can be an address). It may also
    have an initial (reset) value.

    A register is made up of one or more fields. A field starts at a (low)
    first bit and ends as a last (high) bit. The bit range is inclusive.

    The bits in the field can be converted to several types:
        - boolean (one bit only)
        - enum
        - integer (unsigned)

    A register value holds the value of a register.
    The :py:class:`Register` describes a register and the fields it contains.
    The :py:class:`RegisterValue` holds a register value and references a
    :py:class:`Register`.

    Example use
    -----------

    .. code-block:: python

        wdt_ctrla = Register('WDT.CTRLA', 0x0100,
                             RegisterDescription( 8, [
                                 RegisterFieldInt('PERIOD', 0, 3),
                                 RegisterFieldInt('WINDOW', 4, 7)])
                             )
        wdt_status = Register('WDT.STATUS', 0x0101,
                              RegisterDescription(8, [
                                  RegisterFieldBool('SYNCBUSY', 0),
                                  RegisterFieldBool('LOCK', 7)]),
                              0)

        @enum.unique
        class VrefAC0RegSel(enum.IntEnum):
            """"""
            V0_55 = 0x00
            V1_1  = 0x01
            V2_5  = 0x02
            V4_4  = 0x04
            res1  = 0x05
            res2  = 0x06
            AVDD  = 0x07

        vref_ctrla = Register('VREF.CTRLA', 0x00A0,
                              RegisterDescription(8, [
                                  RegisterFieldEnum('AC0REFSEL', 0, 2, VrefAC0RegSel)]),
                              0)

        nvmctrl_addr = Register('NVMCTRL.ADDR', 0x1008,
                                RegisterDescription(16, [
                                    RegisterFieldInt('ADDR', 0, 15, RegisterField.Endian.Big),]),
                                0)
        nvmctrl_addr_value = RegisterValue(nvmctrl_addr, 0x1234)
        nvmctrl_addr_value = RegisterValue(nvmctrl_addr, 0x1234)
        print(f'nvmctrl_addr_value 0x{nvmctrl_addr_value.value:04x} is {nvmctrl_addr_value}')
        nvmctrl_addr_value.value = 0x5678
        print(f'nvmctrl_addr_value 0x{nvmctrl_addr_value.value:04x} is {nvmctrl_addr_value}')

    Output:

    .. code-block:: text

        nvmctrl_addr_value 0x1234 is ADDR:4660
        nvmctrl_addr_value 0x5678 is ADDR:22136

"""
# Postpone type hint evaluation until after the entire class has been parsed.
# This was to become standard in 3.10 but will probably never be.
# @see https://stackoverflow.com/a/33533514
# @see https://mail.python.org/archives/list/python-dev@python.org/thread/CLVXXPQ2T2LQ5MP2Y53VVQFCXYWQJHKZ/
from __future__ import annotations

from collections import OrderedDict
import enum
import sys
from typing import Any, Generator, Type, TypeVar


class RegisterField:
    """Describe a field in a register."""

    @enum.unique
    class Endian(enum.Enum):
        """Endianness of a register."""
        Unknown = '??'
        """The endianness is unknown."""
        Big = 'BE'
        """Values are stored with the most significant byte at the lowest
        address and the least significant byte at the highest address.

        Example ::
            0x12345678 is stored as 0x12 0x34 0x56 0x78.
        """
        Litte = 'LE'
        """Values are stored with the least significant byte at the lowest
        address and the most significant byte at the highest address.

        Example ::
            0x12345678 is stored as 0x78 0x56 0x34 0x12.
        """
        Middle = 'ME'
        """Values are stored as big endian words where each word is stored
        as litte endian.

        Example ::
            0x12345678 is stored as 0x34 0x12 0x78 0x56.
        """

    def __init__(self, name: str,
                 first_bit: int = 0,
                 last_bit: int = 31,
                 conversion: Type = None,
                 endian: Endian = Endian.Unknown):
        """Initialize the instance.

        :param name: The name of the field.
        :param first_bit: The first bit of the register belonging to the field (inclusive).
        :param last_bit: The last bit of the register belonging to the field (inclusive).
        :param TypeVar conversion: The conversion class.
        :param Endian endian: Endianness of the field. This is only relevant for
            fields spanning byte boundaries.
        :raise IndexError: last_bit is smaller than first_bit.
        """
        self._name = name
        if first_bit > last_bit:
            raise IndexError("first bit larger than last bit")
        self._first_bit = first_bit
        self._last_bit = last_bit
        self._conversion = conversion
        self._endian = endian

    def __str__(self) -> str:
        s = f'<{self.__class__.__name__}: {self._name} [{self._first_bit}'
        if self._first_bit == self._last_bit:
            s += '] '
        else:
            s += f':{self.last_bit}] '
        if self._conversion is not None:
            s += f'{self._conversion.__name__}'
        else:
            s += 'None'
        if self._last_bit - self._first_bit > 7:
            s += f' ({self._endian.value})'
        s += '>'
        return s

    @property
    def conversion(self) -> Type:
        """Return class used to convert the value."""
        return self._conversion

    @property
    def endian(self) -> RegisterField.Endian:
        """Endianness of the field."""
        return self._endian

    @property
    def first_bit(self) -> int:
        """First bit of the register containing the field."""
        return self._first_bit

    @property
    def last_bit(self) -> int:
        """Last bit of the register containing the field."""
        return self._last_bit

    @property
    def mask(self) -> int:
        """A mask leaving only bits contained in the field."""
        left = (1 << (self._last_bit + 1)) - 1
        right = (1 << self._first_bit) - 1
        return left & ~right

    @property
    def name(self) -> str:
        """Name of the field."""
        return self._name

    @property
    def type(self) -> Type:
        """Type of the field."""
        return self._conversion

    def convert(self, value: Any) -> Any:
        """Convert the given value to the type of the field."""
        return self._conversion(value)


class RegisterFieldBool(RegisterField):
    """A field in a register that is a single bit."""

    def __init__(self, name: str, bit: int):
        """Initialize the instance.

        :param str name: Name of the field.
        :param int bit: The bit in the register of the field.
        """
        super().__init__(name, bit, bit, bool)


class RegisterFieldInt(RegisterField):
    """A register field containing an integral value."""

    def __init__(self, name: str,
                 first_bit: int,
                 last_bit: int,
                 endian: RegisterField.Endian = RegisterField.Endian.Unknown):
        """Initialize the instance.

        :param str name: Name of the field.
        :param int first_bit: First bit of the register containing the field (inclusive).
        :param int last_bit: Last bit of the register containing the field (inclusive).
        :param RegisterField.Endian endian: Endianness of the field. This is only relevant for
            fields spanning byte boundaries.
        """
        super().__init__(name, first_bit, last_bit, int, endian)
        self._min_value = 0
        self._max_value = (1 << (last_bit + 1 - first_bit)) - 1

    @property
    def min_value(self) -> int:
        """Minimum allowed value in the field (inclusive)."""
        return self._min_value

    @property
    def max_value(self) -> int:
        """Maximum allowed value in the field (inclusive)."""
        return self._max_value


class RegisterFieldEnum(RegisterField):
    def __init__(self, name: str, first_bit: int, last_bit: int, conversion: Type[enum.IntEnum]|Type[enum.IntFlag]):
        """Initialize the instance.

        :param str name: Name of the field.
        :param int first_bit: First bit of the register containing the field (inclusive).
        :param int last_bit: Last bit of the register containing the field (inclusive).
        :param TypeVar conversion: The conversion class which must be an enum.
        """
        super().__init__(name, first_bit, last_bit, conversion)


class RegisterDescription:
    """Describes the layout and fields of a register."""

    def __init__(self, nr_bits: int, fields: list[RegisterField]):
        """Initialize self.

        :param int nr_bits: Number of bits in the register.
        :param list[RegisterField] fields: Fields of the register.
        :raise IndexError: An error was found in the fields.
        :raise KeyError: A field name was used multiple times.
        """
        self._nr_bits = nr_bits
        self._fields_by_name = OrderedDict([(f.name, f) for f in sorted(fields, key=lambda field: field.name)])
        self._fields_by_start = OrderedDict([(f.first_bit, f) for f in sorted(fields, key=lambda field: field.first_bit)])
        # Make sure fields are sorted by rising first_bit.
        self._fields = self._fields_by_start.values()
        if len(self._fields) != len(self._fields_by_name):
            raise KeyError('field name multiply defined')
        if len(self._fields) != len(self._fields_by_start):
            raise KeyError('field position multiply defined')
        self.check_field_consistency()

    def __str__(self) -> str:
        """Convert instance to a human-readable string.

            :rtype: str
        """
        s = f'<{self.__class__.__name__} nr_bits:{self._nr_bits}' +\
            f' fields:['
        first = True
        for first_bit in sorted(self._fields_by_start.keys()):
            field = self._fields_by_start[first_bit]
            if not first:
                s += f', {field}'
            else:
                s += f'{field}'
                first = False
        s += ']>'
        return s

    def check_field_consistency(self):
        """Check that the fields are consistently defined.

        :raise IndexError: An error was found in the fields.
        """
        nr_bits = 0
        bit_offset = -1
        for field in self._fields:
            if bit_offset >= field.first_bit:
                raise IndexError('bits multiply defined')
            nr_bits += field.last_bit - field.first_bit + 1
            if nr_bits > self._nr_bits:
                raise IndexError('too many bits')

    def field(self, name: str) -> RegisterField:
        """Return the field with the given name.

            :param str name: Name of the field.
            :raise KeyError: The given name is not a field.
        """
        return self._fields_by_name[name]

    def fields_by_start_bit(self) -> Generator[RegisterField, None, None]:
        """Yield all register fields in rising order of their start bit.

            :return: A register field per call.
            :rtype: RegisterField
        """
        for field in self._fields:
            yield field

    def fields_by_end_bit(self) -> Generator[RegisterField, None, None]:
        """Yield all register fields in falling order of their end bit.

            :return: A register field per call.
            :rtype: RegisterField
        """
        for field in reversed(self._fields):
            yield field

    def mask_keep_field(self, field: RegisterField) -> int:
        """Create a mask to remove all bits not belonging to the field.

           :param RegisterField field: The field to mask.
           :return: A mask leaving only the bits belonging to the field.
        """
        return field.mask

    def mask_remove_field(self, field: RegisterField) -> int:
        """Create a mask to remove all bits belonging to the field.

            :param RegisterField field: The field to mask.
            :return: A mask removing all bits belonging to the field.
        """
        return ~field.mask

    def value_of_field(self, field: RegisterField, value: int) -> Any:
        """Return the value of the field in the register."""
        v = (value & field.mask) >> field.first_bit
        return field.convert(v)


class Register:
    """A register at an offset with a name and description."""

    def __init__(self, name: str,
                 offset: int,
                 description: RegisterDescription,
                 initial: int = -1):
        """Initialize self.

        :param str name: Name of the register.
        :param int offset: Offset of the register.
        :param RegisterDescription description: Description of the register.
        :param int initial: Initial value of the register after reset.
            Use -1 for unknown initial value.
        """
        self._name = name
        self._offset = offset
        self._description = description
        self._initial = initial

    def __str__(self) -> str:
        """Convert instance to a human-readable string.

            :rtype: str
        """
        s = f'<{self.__class__.__name__}:{self._name}' + \
            f' offset:0x{self._offset:0x} {self._description}>'
        return s

    @property
    def name(self) -> str:
        """Return the register name."""
        return self._name

    @property
    def offset(self) -> int:
        """Offset of the register in memory."""
        return self._offset

    @property
    def description(self) -> RegisterDescription:
        """Description of the register."""
        return self._description

    @property
    def initial(self) -> int:
        """Initial value of the register after reset. -1 is unknown."""
        return self._initial


class RegisterValue:
    def __init__(self, register: Register,
                 value: int):
        self.register = register
        self.value = value

    def __str__(self) -> str:
        """Convert instance to a human-readable string."""
        s = ''
        for field in self.register.description.fields_by_start_bit():
            v = self.register.description.value_of_field(field, self.value)
            if isinstance(v, enum.Enum):
                v = v.name
            s += f'{field.name}:{v} '
        return s.rstrip()

    def __getattr__(self, name: str) -> bool|int|enum.IntEnum|enum.IntFlag:
        """Get the value of the named field in the register.

        :param str name: Name of the field.
        :rtype: bool|int|enum.IntEnum|enum.IntFlag
        :return: Value of the field in the register.
        :raise AttributeError: the field does not exist.
        """
        the_field = self.register.description.field(name)
        value = (self.value & the_field.mask) >> the_field.first_bit
        return the_field.convert(value)

    def set(self, name: str, value: bool|int|enum.IntEnum|enum.IntFlag):
        """Set the value of the named field in the register.

        :param str name: Name of the field.
        :param bool|int|enum.IntEnum|enum.IntFlag value: Value of the field in the register.
        :raise AttributeError: the field does not exist.
        """
        the_field = self.register.description.field(name)
        v = int(value) << the_field.first_bit
        self._value = (self.value & ~the_field.mask) | v


def example():
    """Module demonstration."""
    wdt_ctrla = Register('WDT CTRLA', 0x0100,
                         RegisterDescription( 8, [
                             RegisterFieldInt('PERIOD', 0, 3),
                             RegisterFieldInt('WINDOW', 4, 7)])
                         )
    print(f'wdt_ctrla contains integer fields: {wdt_ctrla}')
    wdt_status = Register('WDT STATUS', 0x0101,
                          RegisterDescription(8, [
                              RegisterFieldBool('SYNCBUSY', 0),
                              RegisterFieldBool('LOCK', 7)]),
                          0)
    print(f'wdt_status contains boolean fields: {wdt_status}')

    @enum.unique
    class VrefAC0RegSel(enum.IntEnum):
        """"""
        V0_55 = 0x00
        """0.55V"""
        V1_1  = 0x01
        """1.1V"""
        V2_5  = 0x02
        """2.5V"""
        V4_4  = 0x04
        """4.4V"""
        res1  = 0x05
        """Reserved"""
        res2  = 0x06
        """Reserved"""
        AVDD  = 0x07
        """AVDD"""

    vref_ctrla = Register('VREF CTRLA', 0x00A0,
                          RegisterDescription(8, [
                              RegisterFieldEnum('AC0REFSEL', 0, 2, VrefAC0RegSel)]),
                          0)
    print(f'vref_ctrla contains enum fields: {vref_ctrla}')

    @enum.unique
    class NvmectrlCtrlACmd(enum.IntEnum):
        NoCommand = 0x00
        """No command."""
        WP = 0x01
        """Write page buffer to memory (NVMCTRL.ADDR selects which memory)."""
        EP = 0x02
        """Erase page buffer (NVMCTRL.ADDR selects which memory)."""
        ERWP = 0x03
        """Erase and write page buffer (NVMCTRL.ADDR selects which memory)."""
        PBC = 0x04
        """Page buffer clear."""
        CHER = 0x05
        """Chip erase: erase Flash and EEPROM (unless EESAVE in FUSE.SYSCFG is '1')."""
        EEER = 0x06
        """EEPROM Erase."""
        WFU = 0x07
        """Write fuse (only accessible through UPDI)."""
    nvmctrl_ctrla = Register('NVMCTRL CTRLA', 0x1000,
                              RegisterDescription(8, [
                                  RegisterFieldEnum('CTRLA', 0, 2, NvmectrlCtrlACmd),
                              ]),
                             0)
    print(f'nvmctrl_ctrla: {nvmctrl_ctrla}')
    nvmctrl_ctrla_value = RegisterValue(nvmctrl_ctrla, 0x03)
    print(f'nvmctrl_ctrla_value 0x{nvmctrl_ctrla_value.value:02x} is {nvmctrl_ctrla_value}')
    nvmctrl_addr = Register('NVMCTRL ADDR', 0x1008,
                            RegisterDescription(16, [
                                RegisterFieldInt('ADDR', 0, 15, RegisterField.Endian.Big),]),
                            0)
    print(f'nvmctrl_addr: {nvmctrl_addr}')
    nvmctrl_addr_value = RegisterValue(nvmctrl_addr, 0x1234)
    print(f'nvmctrl_addr_value 0x{nvmctrl_addr_value.value:04x} is {nvmctrl_addr_value}')
    nvmctrl_addr_value.value = 0x5678
    print(f'nvmctrl_addr_value 0x{nvmctrl_addr_value.value:04x} is {nvmctrl_addr_value}')

    wdt_ctrla_value = RegisterValue(wdt_ctrla, 0x73)
    print(f'wdt_ctrla_value 0x{wdt_ctrla_value.value:02x} is {wdt_ctrla_value}')
    wdt_status_value = RegisterValue(wdt_status, 0x80)
    print(f'wdt_status_value 0x{wdt_status_value.value:02x} is {wdt_status_value}')
    vref_ctrla_value = RegisterValue(vref_ctrla, 0x02)
    print(f'vref_ctrla_value 0x{vref_ctrla_value.value:02x} is {vref_ctrla_value}')
    vref_ctrla_value.value = 0x07
    print(f'vref_ctrla_value 0x{vref_ctrla_value.value:02x} is {vref_ctrla_value}')


if __name__ == '__main__':
    if sys.version_info < (3, 10):
        print('FATAL ERROR: Python 3.10.x or later is required.')
        sys.exit(1)
    example()



#!/usr/bin/env python3
# vim: fileencoding=utf-8 ts=4
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: © Copyright 2024 by Christian Dönges. All rights reserved.
# SPDXID: SPDXRef-xmlutils-py
"""XML handling utilities."""

import logging
import xml.etree.ElementTree as ET

try:
    from lxml import etree as ET
    _USING_LXM = True
except ImportError:
    import xml.etree.ElementTree as ET
    _USING_LXM = False


def elements_equal(element1: ET.Element | ET.ElementTree,
                   element2: ET.Element | ET.ElementTree,
                   logger: logging.Logger = logging.getLogger(__name__)) -> bool:
    """Check if two XML elements are equal.

    Compares the tags, attributes, children (recursively), tail, and text of
    both *elements*, not their textual representations.

    :param xml.etree.ElementTree.Element element1: First XML element to compare.
    :param xml.etree.ElementTree.Element element2: Second XML element to compare.
    :param logging.Logger logger: Logger for diagnostics.
    :rtype: bool
    :return: True if elements are equal, False otherwise.

    .. versionadded:: 1.6.0
    """
    if isinstance(element1, ET.ElementTree):
        element1 = element1.getroot()
        logger.debug('Converted element <%s> from ElementTree to Element for comparison.', element1.tag)
    if isinstance(element2, ET.ElementTree):
        element2 = element2.getroot()
        logger.debug('Converted element <%s> from ElementTree to Element for comparison.', element2.tag)

    if element1.tag != element2.tag:
        logger.debug('Tag "%s" != "%s"', element1.tag, element2.tag)
        return False
    if element1.text != element2.text:
        logger.debug('Text "%s" != "%s"', element1.text, element2.text)
        return False
    if element1.tail != element2.tail:
        logger.debug('Tail "%s" != "%s"', element1.tail, element2.tail)
        return False
    if element1.attrib != element2.attrib:
        logger.debug('Attribute "%s" != "%s"', element1.attrib, element2.attrib)
        return False
    if len(element1) != len(element2):
        logger.debug('Length %d != %d', len(element1), len(element2))
        return False
    return all(elements_equal(child1, child2) for child1, child2 in zip(element1, element2))


def log_xml(fn_logger: logging.Logger,
            xml_elem: ET.Element | ET.ElementTree,
            level: int = logging.DEBUG,
            indent: int = 0,
            indent_level_start: int = 0,
            indent_spaces: int = 2):
    """Log the XML as one or more log entries.

    :param logging.Logger fn_logger: The logger to log to.
    :param etree._Element xml_elem: The XML element to log.
    :param int level: Level for logging (e.g logging.DEBUG, logging.INFO, ...)
    :param int indent: Number of spaces to indent all entries. This does not
        change the indentation added for nested elements (indent_spaces spaces per level).
    :param int indent_level_start: Start indentation at this level.
        If > 0, the parameter "indent" is ignored.
    :param int indent_spaces: Number of spaces to indent per level.

    .. note:: ``indent`` is the old interface for indenting the entire
        ``xml_element``. In order to be closer to the interface of the Python
        standard library, ``indent_level_start`` and ``indent_spaces`` have been
        introduced. Since there is no compatible way to deprecate ``indent``,
        it will stay but the recommendation is to switch to the new parameters.

    .. versionchanged:: 1.6.0 Moved here from :py:func:`pymisclib.utilities.log_xml`.
    .. versionchanged:: 1.6.2 Formatting improved.
    .. versionchanged:: 1.7.1 Added parameters ``indent_level_start`` and ``indent_spaces``.
    """
    if not fn_logger.isEnabledFor(level):
        # Save lots of processing cycles.
        return

    if indent < 0:
        raise ValueError('indent must be >= 0')
    if indent_level_start < 0:
        raise ValueError('indent_level_start must be >= 0')
    if indent_spaces < 0:
        raise ValueError('indent_space must be >= 0')

    if isinstance(xml_elem, ET.ElementTree):
        xml_elem = xml_elem.getroot()

    # Try to convert indent to an indent_level_start.
    if indent > 0 and indent_level_start == 0:
        if indent_spaces > 0:
            if indent % indent_spaces == 0:
                indent_level_start = int(indent / indent_spaces)
                indent = 0

    try:
        ET.indent(xml_elem, space = ' ' * indent_spaces, level = indent_level_start)
    except Exception as e:
        fn_logger.error('Failed to indent XML element: %s', e)
        fn_logger.debug('    XML element: %s', xml_elem)
        return
    try:
        if _USING_LXM:
            s = ET.tostring(xml_elem, pretty_print=True, encoding='unicode')
        else:
            s = ET.tostring(xml_elem, encoding='unicode')
            # xml.etree.ElementTree.indent() does not correctly indent the first line.
            if indent_level_start > 0:
                s = ' ' * (indent_spaces * indent_level_start) + s
    except Exception as e:
        fn_logger.error('Failed to convert XML element to string: %s', e)
        fn_logger.debug('    XML element: %s', xml_elem)

    if indent == 0:
        fmt = '%s'
    else:
        fmt = f'{" " * indent}%s'
    for line in s.splitlines():
        fn_logger.log(level, fmt, line)



#!/usr/bin/env python3
# vim: fileencoding=utf-8 ts=4
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: © Copyright 2024 by Christian Dönges. All rights reserved.
# SPDXID: SPDXRef-test-xmlutils-py

from xml.etree import ElementTree as ET

from pymisclib.xmlutils import elements_equal


def test_flat_equal():
    e1 = ET.Element('element')
    e2 = ET.Element('element')
    assert elements_equal(e1, e2)


def test_flat_not_equal():
    e1 = ET.Element('element_one')
    e2 = ET.Element('element_two')
    assert not elements_equal(e1, e2)


def test_flat_with_attributes_equal():
    e1 = ET.Element('element',
                    attrib={'a': 'val1', 'b': 'val2'})
    e2 = ET.Element('element',
                    attrib={'a': 'val1', 'b': 'val2'})
    assert elements_equal(e1, e2)

    e3 = ET.Element('element',
                    attrib={'a': 'val1', 'b': 'val2'})
    e4 = ET.Element('element',
                    attrib={'b': 'val2', 'a': 'val1'})
    assert elements_equal(e3, e4)


    e3 = ET.Element('element',
                    attrib={'a': 'val1', 'b': 'val2',
        'c': 'val3'})
    e4 = ET.Element('element',
                    attrib={'c': 'val3', 'b': 'val2',
        'a': 'val1'})
    assert elements_equal(e3, e4)


def test_flat_with_attributes_not_equal():
    """Compare flat elements with different attributes."""
    # Attribute value differs.
    e1 = ET.Element('element',
                    attrib={'a': 'val1'})
    e2 = ET.Element('element',
                    attrib={'a': 'val2'})
    assert not elements_equal(e1, e2)

    # Number of attributes differs.
    e3 = ET.Element('element',
                    attrib={'a': 'val1', 'b': 'val2'})
    assert not elements_equal(e1, e3)

    # Wrong value type.
    e4 = ET.Element('element',
                    attrib={'a': 'val1', 'b': '2'})
    e5 = ET.Element('element',
                    attrib={'a': 'val1', 'b': 2})
    assert not elements_equal(e4, e5)


def test_flat_with_text_equal():
    """Compare flat elements with the same text."""
    e1 = ET.Element('element', text='text1')
    e2 = ET.Element('element', text='text1')
    assert elements_equal(e1, e2)


def test_flat_with_text_not_equal():
    """Compare flat elements with different text."""
    # Different text
    e1 = ET.Element('element', text='text1')
    e2 = ET.Element('element', text='text2')
    assert not elements_equal(e1, e2)

    # Case is significant.
    e3 = ET.Element('element', text='Text1')
    assert not elements_equal(e1, e3)
    e4 = ET.Element('element', text='teXt1')
    assert not elements_equal(e1, e4)


def test_flat_with_tail_equal():
    """Compare flat elements with the same tail."""
    e1 = ET.Element('element', tail='tail1')
    e2 = ET.Element('element', tail='tail1')
    assert elements_equal(e1, e2)

    # Non-ASCII characters.
    e3 = ET.Element('element', tail='This is a Ⲧäïл.')
    e4 = ET.Element('element', tail='This is a Ⲧäïл.')
    assert elements_equal(e3, e4)


def test_flat_with_tail_not_equal():
    """Compare flat elements with different tail."""
    e1 = ET.Element('element', tail='tail1')
    e2 = ET.Element('element', tail='tail2')
    assert not elements_equal(e1, e2)

    # Case is significant.
    e3 = ET.Element('element', tail='Tail1')
    assert not elements_equal(e1, e3)
    e4 = ET.Element('element', tail='taIL1')
    assert not elements_equal(e1, e4)

    # Whitespace is significant.
    e5 = ET.Element('element', tail='This is a tail.')
    e6 = ET.Element('element', tail='This  is a tail.')
    assert not elements_equal(e5, e6)

    # Non-ASCII characters.
    e7 = ET.Element('element', tail='This is a Täïл.')
    e8 = ET.Element('element', tail='This is a Ⲧäïл.')
    assert not elements_equal(e7, e8)


def test_flat_all_equal():
    """Compare flat elements with attributes, tail, and text."""
    e1 = ET.Element('element',
                    attrib={'a': 'val1', 'b': 'val2'},
                    text='text1',
                    tail='tail1')
    e2 = ET.Element('element',
                    attrib={'b': 'val2', 'a': 'val1'},
                    text='text1',
                    tail='tail1')
    assert elements_equal(e1, e2)


def test_children_equal_1():
    """Compare elements with identical one child."""
    e1 = ET.Element('element')
    c1 = ET.SubElement(e1, 'child',
                       attrib={'a': 'val1', 'b': 'val2'},
                       tail='tail1',
                       text='text1')
    e2 = ET.Element('element')
    c2 = ET.SubElement(e2, 'child',
                       attrib={'b': 'val2', 'a': 'val1'},
                       tail='tail1',
                       text='text1')
    assert elements_equal(e1, e2)


def test_children_order_equal_2():
    """Compare elements with identical two children."""
    e1 = ET.Element('parent')
    c11 = ET.SubElement(e1, 'child',
                       attrib={'a': 'val1', 'b': 'val2'},
                       tail='tail1',
                       text='text1')
    c12 = ET.SubElement(e1, 'child',
                       attrib={'c': 'val1', 'd': 'val2'},
                       tail='tail2',
                       text='text2')
    e2 = ET.Element('parent')
    c21 = ET.SubElement(e2, 'child',
                        attrib={'b': 'val2', 'a': 'val1'},
                        tail='tail1',
                        text='text1')
    c22 = ET.SubElement(e2, 'child',
                        attrib={'d': 'val2', 'c': 'val1'},
                        tail='tail2',
                        text='text2'),
    assert elements_equal(e1, e2)


def test_children_order_not_equal_2():
    """Compare elements with identical two children in wrong order."""
    e1 = ET.Element('parent')
    c11 = ET.SubElement(e1, 'child',
                       attrib={'a': 'val1', 'b': 'val2'},
                       tail='tail1',
                       text='text1')
    c12 = ET.SubElement(e1, 'child',
                       attrib={'c': 'val1', 'd': 'val2'},
                       tail='tail2',
                       text='text2')
    e2 = ET.Element('parent')
    c21 = ET.SubElement(e2, 'child',
                        attrib={'d': 'val2', 'c': 'val1'},
                        tail='tail2',
                        text='text2'),
    c22 = ET.SubElement(e2, 'child',
                        attrib={'b': 'val2', 'a': 'val1'},
                        tail='tail1',
                        text='text1')
    assert not elements_equal(e1, e2)




#!/usr/bin/env python3
# vim: fileencoding=utf8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2023 by Christian Dönges
# SPDXID: SPDXRef-ansi-terminal-py
"""Handle console operations on an ANSI-compliant terminal.

.. see: https://en.wikipedia.org/wiki/ANSI_escape_code

.. deprecated:: 1.6.0 Will be removed in 2.0.

"""
import enum
import logging
import os
import re
import sys
from dataclasses import dataclass
from math import log2
from time import sleep


if os.name != 'nt':
    # termios is a POSIX API not available on Windows.
    import termios
    import tty
else:
    from ctypes import byref, c_int, windll
    import msvcrt


@enum.unique
class BgColor(enum.Enum):
    """Background colors and the ANSI format strings used to create them."""
    Black = '\x1B[40m'
    Red = '\x1B[41m'
    Green = '\x1B[42m'
    Yellow = '\x1B[43m'
    Blue = '\x1B[44m'
    Magenta = '\x1B[45m'
    Cyan = '\x1B[46m'
    White = '\x1B[47m'
    BrightBlack = '\x1B[100m'
    BrightRed = '\x1B[101m'
    BrightGreen = '\x1B[102m'
    BrightYellow = '\x1B[103m'
    BrightBlue = '\x1B[104m'
    BrightMagenta = '\x1B[105m'
    BrightCyan = '\x1B[106m'
    BrightWhite = '\x1B[107m'


@enum.unique
class FgColor(enum.Enum):
    """Foreground colors and the ANSI format strings used to create them."""
    Black = '\x1B[30m'
    Red = '\x1B[31m'
    Green = '\x1B[32m'
    Yellow = '\x1B[33m'
    Blue = '\x1B[34m'
    Magenta = '\x1B[35m'
    Cyan = '\x1B[36m'
    White = '\x1B[37m'
    BrightBlack = '\x1B[90m'
    BrightRed = '\x1B[91m'
    BrightGreen = '\x1B[92m'
    BrightYellow = '\x1B[93m'
    BrightBlue = '\x1B[94m'
    BrightMagenta = '\x1B[95m'
    BrightCyan = '\x1B[96m'
    BrightWhite = '\x1B[97m'


@dataclass
class AnsiColor:
    """24 bit RGB color for use with ANSI terminals.

    There are three components, each of which must be 0..255:

        - red
        - green
        - blue

    Note that not all terminals support 24 bit colors.
    """
    red: int = 0
    green: int = 0
    blue: int = 0

    @property
    def bg(self) -> str:
        """Return a string to set the background to this color."""
        return f'\x1B[48;2;{self.red};{self.green};{self.blue}m'

    @property
    def fg(self) -> str:
        """Return a string to set the foreground to this color."""
        return f'\x1B[38;2;{self.red};{self.green};{self.blue}m'


class AnsiControl:
    """Popular ANSI control sequences."""

    @staticmethod
    def erase_line():
        """Erase the current line."""
        return '\x1B[2K'

    @staticmethod
    def erase_screen():
        """Erase the entire screen content.

        Depending on the terminal, this may scroll up the entire screen, so the
        screen content will be saved to the scrollback buffer.
        """
        return '\x1B[2J'

    @staticmethod
    def erase_scrollback_buffer():
        """Erase the scrollback buffer without erasing the screen content."""
        return '\x1B[3J'

    @staticmethod
    def erase_to_end_of_line():
        """Erase from the cursor position to the end of the current line."""
        return '\x1B[0K'

    @staticmethod
    def erase_to_start_of_line():
        """Erase from the cursor position to the start of the current line."""
        return '\x1B[1K'

    @staticmethod
    def move_cursor_to(column: int = 1, row: int = 1) -> str:
        """Move the cursor to the given position.

        :param int column: New cursor column (x-position).
        :param int row: New cursor row (y-position).
        :return: Control sequence.
        :rtype str:
        :raise ValueError: The row or column is less than 1.
        """
        if row < 1:
            raise ValueError('row too small')
        if column < 1:
            raise ValueError('column too small')
        return f'\x1b[{row};{column}H'

    @staticmethod
    def move_cursor_back(nr_columns: int = 1) -> str:
        """Move the cursor back by `nr_columns` columns.

        If the cursor is at the edge of the screen, this has no effect.

        :param int nr_columns: Number of columns to move the cursor by.
        :return: Control sequence.
        :rtype str:
        :raise ValueError: The number of columns is less than 1.
        """
        if nr_columns < 1:
            raise ValueError('too few columns')
        return f'\x1b[{nr_columns}D'

    @staticmethod
    def move_cursor_down(nr_rows: int = 1):
        """Move the cursor down by `nr_rows` rows.

        If the cursor is at the edge of the screen, this has no effect.

        :param int nr_rows: Number of rows to move the cursor by.
        :return: Control sequence.
        :rtype str:
        :raise ValueError: The number of rows is less than 1.
        """
        if nr_rows < 1:
            raise ValueError('too few rows')
        return f'\x1b[{nr_rows}B'

    @staticmethod
    def move_cursor_forward(nr_columns: int = 1):
        """Move the cursor forward by `nr_columns` columns.

        If the cursor is at the edge of the screen, this has no effect.

        :param int nr_columns: Number of columns to move the cursor by.
        :return: Control sequence.
        :rtype str:
        :raise ValueError: The number of columns is less than 1.
        """
        if nr_columns < 1:
            raise ValueError('too few columns')
        return f'\x1b[{nr_columns}C'

    @staticmethod
    def move_cursor_up(nr_rows: int = 1):
        """Move the cursor up by `nr_rows` rows.

        If the cursor is at the edge of the screen, this has no effect.

        :param int nr_rows: Number of rows to move the cursor by.
        :return: Control sequence.
        :rtype str:
        :raise ValueError: The number of rows is less than 1.
        """
        if nr_rows < 1:
            raise ValueError('too few rows')
        return f'\x1b[{nr_rows}A'

    @staticmethod
    def scroll_page_down(nr_pages: int = 1):
        """Scroll down by nr_pages pages/screens.

        :param int nr_pages: Number of pages to scroll down by.
        :return: Control sequence.
        :rtype str:
        :raise ValueError: nr_pages is < 1
        """
        if nr_pages < 1:
            raise ValueError('scroll down at least one page')
        return f'\x1B{nr_pages}T'

    @staticmethod
    def scroll_page_up(nr_pages: int = 1):
        """Scroll up by nr_pages pages/screens.

        :param int nr_pages: Number of pages to scroll up by.
        :return: Control sequence.
        :rtype str:
        :raise ValueError: nr_pages is < 1
        """
        if nr_pages < 1:
            raise ValueError('scroll up at least one page')
        return f'\x1B{nr_pages}S'


@enum.unique
class TextStyle(enum.Enum):
    """Text styles and the ANSI format strings used to create them."""
    Reset = '\x1B[0m'  # All attributes off.
    Bold = '\x1B[1m'
    Faint = '\x1B[2m'
    Italic = '\x1B[3m'
    Underline = '\x1B[4m'
    SlowBlink = '\x1B[5m'  # Text blinks less than 150 times per minute.
    RapidBlink = '\x1B[6m'  # Text blinks more than 150 times per minute. Not widely supported.
    Reverse = '\x1B[7m'  # Swaps foreground and background colors. Inconsistently implemented.
    Hide = '\x1B[8m'  # Not widely supported.
    StrikeOut = '\x1B[9m'  # Characters marked for deltetion. Not supported in Terminal.app.
    PrimaryFont = '\x1B[10m'
    AlternateFont1 = '\x1B[11m'
    AlternateFont2 = '\x1B[12m'
    AlternateFont3 = '\x1B[13m'
    AlternateFont4 = '\x1B[14m'
    AlternateFont5 = '\x1B[15m'
    AlternateFont6 = '\x1B[16m'
    AlternateFont7 = '\x1B[17m'
    AlternateFont8 = '\x1B[18m'
    AlternateFont9 = '\x1B[19m'
    GothicFont = '\x1B[20m'  # Rarely supprorted.
    DoubleUnderline = '\x1B[21m'  # Often NoBold.
    NormalIntensity = '\x1B[22m'  # Neither Bold nor Faint.
    NoItalic = '\x1B[23m'
    NoUnderline = '\x1B[24m'
    NoBlink = '\x1B[25m'
    NoReverse = '\x1B[27m'
    NoHide = '\x1B[28m'
    NoStrikeOut = '\x1B[29m'
    Frame = '\x1B[51m'
    Encircle = '\x1B[52m'
    Overline = '\x1B[53m'
    Superscript = '\x1B[73m'
    Subscript = '\x1B[74m'
    NoSuperSub = '\x1B[75m'  # Neither Superscript nor Subscript.


@dataclass
class ColsRows:
    """A column and row position or size."""
    column: int
    row: int

    def __str__(self) -> str:
        """Return a string representation of the instance."""
        return f'({self.column};{self.row})'


def enable_vt100_mode_on_windows():
    """If on Windows, enable VT100 mode to allow processing of ANSI codes.

        :doc: https://bugs.python.org/issue29059
    """
    if os.name != 'nt':
        return
    stdout_handle = windll.kernel32.GetStdHandle(c_int(-11))  # stdout is 11 on windows
    old_mode = c_int(0)
    windll.kernel32.GetConsoleMode(c_int(stdout_handle), byref(old_mode))
    new_mode = c_int(old_mode.value | 4)  # ENABLE_VIRTUAL_TERMINAL_PROCESSING
    windll.kernel32.SetConsoleMode(c_int(stdout_handle), new_mode)


@dataclass
class AnsiTerminal:
    full: bool = False  # True: full cursor control
    autoflush: bool = False  # automatically flush
    _cursor_column: int = -1
    _cursor_row: int = -1
    _fin = sys.stdin.fileno()
    _fout = sys.stdout.fileno()
    _have_posix: bool = (os.name != 'nt')  # Running on a POSIX system (True) or Windows (False)?
    _logger: logging.Logger = logging.getLogger(__name__)
    _terminal_columns: int = -1
    _terminal_rows: int = -1

    def __post_init__(self):
        if not sys.stdin.isatty():
            self._logger.error('stdin is not a TTY.')
            raise EnvironmentError('stdin not a TTY')
        if not sys.stdout.isatty():
            self._logger.error('stdout is not a TTY.')
            raise EnvironmentError('stdout not a TTY')
        enable_vt100_mode_on_windows()
        if self.full and not sys.stdin.readable():
            self._logger.warning('stdin is not readable, running without full mode.')
            self.full = False
        self._determine_current_cursor()
        (self._terminal_columns, self._terminal_rows) = os.get_terminal_size(self._fout)

    def _determine_current_cursor(self):
        """Determine the current cursor position."""
        self._logger.debug('%s._determine_current_cursor()', self.__class__.__name__)
        if not self.full:
            self._logger.debug('Current cursor [fake]: (1;1)')
            self._cursor_column = 1
            self._cursor_row = 1
            return

        if self._have_posix:
            reply = self._get_current_cursor_posix()
        else:
            reply = self._get_current_cursor_windows()

        m = re.match(r'\x1b\[(?P<row>\d+);(?P<col>\d+)R', reply)
        if m is None:
            self._logger.error('Expected DSR response "ESC[row;colR" got "%s".',
                               reply.encode('utf-8').hex())
            raise ValueError('unexpected DSR response')
        self._cursor_row = int(m.group('row'))
        self._cursor_column = int(m.group('col'))
        self._logger.debug('Current cursor: (%d;%d)', self._cursor_column, self._cursor_row)

    def _get_current_cursor_posix(self) -> str:
        """Return a string containing the current cursor position on POSIX."""
        tattr = termios.tcgetattr(self._fin)
        reply = ''
        try:
            # Switch to CBreak mode.
            tty.setcbreak(self._fin, termios.TCSANOW)
            sys.stdout.write('\x1b[6n')  # DSR - Display Status Request
            sys.stdout.flush()

            while True:
                reply += sys.stdin.read(1)
                if reply[-1] == 'R':
                    break
        except RuntimeError as e:
            self._logger.error('While reading: %s', e)
        finally:
            # Restore previous mode.
            termios.tcsetattr(self._fin, termios.TCSANOW, tattr)

        return reply

    def _get_current_cursor_windows(self) -> str:
        """Return a string containing the current cursor position on Windows.

            :doc: https://stackoverflow.com/a/69006335
        """
        reply = ''
        try:
            sys.stdout.write('\x1b[6n')  # DSR - Display Status Request
            sys.stdout.flush()

            input_pending = True
            while input_pending:
                # Read a single keypress from the keyboard input buffer.
                reply += msvcrt.getch().decode('ascii')
                # Is there more in the input buffer?
                input_pending = msvcrt.kbhit()
        except RuntimeError as e:
            self._logger.error('While reading: %s', e)

        return reply

    def _update_cursor_position(self, column_delta: int, row_delta: int):
        """Update the cached cursor position with the given delta.

           The cursor position is bounded by the screen, so it is always
           valid.

           :param int column_delta: Number of columns the cursor was moved
               left (negative) or right (positive).
           :param int row_delta: Number of rows the cursor was moded
                up (negative) or down (positive).
        """
        self._logger.debug('_update_cursor_position(%d, %d)', column_delta, row_delta)
        self._cursor_column += column_delta
        if self._cursor_column < 1:
            self._cursor_column = 1
        elif self._cursor_column > self._terminal_columns:
            self._cursor_column = self._terminal_columns

        self._cursor_row += row_delta
        if self._cursor_row < 1:
            self._cursor_row = 1
        elif self._cursor_row > self._terminal_rows:
            self._cursor_row = self._terminal_rows

    def cursor_position(self, use_cached: bool = False) -> ColsRows:
        """Return the current cursor position.

        :param bool use_cached: If `True`, the positioned cached by the instance
            is used. This is faster than `False`, which queries the terminal,
            but could be wrong.
        :return: Current position of the cursor.
        :rtype: ColsRows
        """
        if not use_cached:
            self._determine_current_cursor()
        return ColsRows(column=self._cursor_column, row=self._cursor_row)

    def erase_line(self):
        """Erase the current line."""
        self._logger.debug('erase_line()')
        sys.stdout.write(AnsiControl.erase_line())
        if self.autoflush:
            sys.stdout.flush()

    def erase_screen(self):
        """Erase the entire screen content.

        Depending on the terminal, this may scroll up the entire screen, so the
        screen content will be saved to the scrollback buffer.
        """
        self._logger.debug('erase_screen()')
        sys.stdout.write(AnsiControl.erase_screen())
        if self.autoflush:
            sys.stdout.flush()

    def erase_scrollback_buffer(self):
        """Erase the scrollback buffer without erasing the screen content."""
        self._logger.debug('erase_scrollback_buffer()')
        sys.stdout.write(AnsiControl.erase_scrollback_buffer())
        if self.autoflush:
            sys.stdout.flush()


    def erase_to_end_of_line(self):
        """Erase from the cursor position to the end of the current line."""
        self._logger.debug('erase_to_end_of_line()')
        sys.stdout.write(AnsiControl.erase_to_end_of_line())
        if self.autoflush:
            sys.stdout.flush()

    def erase_to_start_of_line(self):
        """Erase from the cursor position to the start of the current line."""
        self._logger.debug('erase_to_start_of_line()')
        sys.stdout.write(AnsiControl.erase_to_start_of_line())
        if self.autoflush:
            sys.stdout.flush()

    def move_cursor_to(self, column: int = 1, row: int = 1):
        """Move the cursor to the given position.

        :param int column: New cursor column (x-position).
        :param int row: New cursor row (y-position).
        """
        self._logger.debug('move_cursor_to(%d, %d)', column, row)
        if column > self._terminal_columns or column <= 0:
            raise ValueError('column out of range')
        if row > self._terminal_rows or row <= 0:
            raise ValueError('row out of range')
        sys.stdout.write(AnsiControl.move_cursor_to(column=column, row=row))
        self._cursor_column = column
        self._cursor_row = row
        if self.autoflush:
            sys.stdout.flush()

    def move_cursor_back(self, nr_columns: int = 1):
        """Move the cursor back by `nr_columns` columns.

        If the cursor is at the edge of the screen, this has no effect.

        :param int nr_columns: Number of columns to move the cursor by.
        """
        self._logger.debug('move_cursor_back(%d)', nr_columns)
        if nr_columns < 1:
            raise ValueError('too few rows')
        sys.stdout.write(AnsiControl.move_cursor_back(nr_columns))
        self._update_cursor_position(-nr_columns, 0)
        if self.autoflush:
            sys.stdout.flush()

    def move_cursor_down(self, nr_rows: int = 1):
        """Move the cursor down by `nr_rows` rows.

        If the cursor is at the edge of the screen, this has no effect.

        :param int nr_rows: Number of rows to move the cursor by.
        """
        self._logger.debug('move_cursor_down(%d)', nr_rows)
        if nr_rows < 1:
            raise ValueError('too few rows')
        sys.stdout.write(AnsiControl.move_cursor_down(nr_rows))
        self._update_cursor_position(0, nr_rows)
        if self.autoflush:
            sys.stdout.flush()

    def move_cursor_forward(self, nr_columns: int = 1):
        """Move the cursor forward by `nr_columns` columns.

        If the cursor is at the edge of the screen, this has no effect.

        :param int nr_columns: Number of columns to move the cursor by.
        """
        self._logger.debug('move_cursor_forward(%d)', nr_columns)
        if nr_columns < 1:
            raise ValueError('too few rows')
        sys.stdout.write(AnsiControl.move_cursor_forward(nr_columns))
        self._update_cursor_position(nr_columns, 0)
        if self.autoflush:
            sys.stdout.flush()

    def move_cursor_up(self, nr_rows: int = 1):
        """Move the cursor up by `nr_rows` rows.

        If the cursor is at the edge of the screen, this has no effect.

        :param int nr_rows: Number of rows to move the cursor by.
        """
        self._logger.debug('move_cursor_up(%d)', nr_rows)
        if nr_rows < 1:
            raise ValueError('too few rows')
        sys.stdout.write(AnsiControl.move_cursor_up(nr_rows))
        self._update_cursor_position(0, -nr_rows)
        if self.autoflush:
            sys.stdout.flush()

    def scroll_page_down(self, nr_pages: int = 1):
        """Scroll down by nr_pages pages/screens.

        :param int nr_pages: Number of pages to scroll down by.
        :raise ValueError: nr_pages is < 1
        """
        self._logger.debug('scroll_page_down(%d)', nr_pages)
        if nr_pages < 1:
            raise ValueError('scroll down at least one page')
        sys.stdout.write(AnsiControl.scroll_page_down(nr_pages))
        if self.autoflush:
            sys.stdout.flush()

    def scroll_page_up(self, nr_pages: int = 1):
        """Scroll up by nr_pages pages/screens.

        :param int nr_pages: Number of pages to scroll up by.
        :raise ValueError: nr_pages is < 1
        """
        self._logger.debug('scroll_page_up(%d)', nr_pages)
        if nr_pages < 1:
            raise ValueError('scroll up at least one page')
        sys.stdout.write(AnsiControl.scroll_page_up(nr_pages))
        if self.autoflush:
            sys.stdout.flush()

    @property
    def terminal_size(self) -> ColsRows:
        """Number of columns and rows in the terminal.

        :return: Number of columns and rows of the terminal.
        :rtype ColsRows:
        """
        return ColsRows(self._terminal_columns, self._terminal_rows)


def showcase_3bit_colors():
    """Print a color chart."""
    print(f'''{TextStyle.Reset.value}Classic Color Chart
                   Normal                  Bright
    Black   {FgColor.Black.value}Foreground{TextStyle.Reset.value} {BgColor.Black.value}Background{TextStyle.Reset.value}   {FgColor.BrightBlack.value}Foreground{TextStyle.Reset.value} {BgColor.BrightBlack.value}Background{TextStyle.Reset.value}
    Red     {FgColor.Red.value}Foreground{TextStyle.Reset.value} {BgColor.Red.value}Background{TextStyle.Reset.value}   {FgColor.BrightRed.value}Foreground{TextStyle.Reset.value} {BgColor.BrightRed.value}Background{TextStyle.Reset.value}
    Green   {FgColor.Green.value}Foreground{TextStyle.Reset.value} {BgColor.Green.value}Background{TextStyle.Reset.value}   {FgColor.BrightGreen.value}Foreground{TextStyle.Reset.value} {BgColor.BrightGreen.value}Background{TextStyle.Reset.value}
    Yellow  {FgColor.Yellow.value}Foreground{TextStyle.Reset.value} {BgColor.Yellow.value}Background{TextStyle.Reset.value}   {FgColor.BrightYellow.value}Foreground{TextStyle.Reset.value} {BgColor.BrightYellow.value}Background{TextStyle.Reset.value}
    Blue    {FgColor.Blue.value}Foreground{TextStyle.Reset.value} {BgColor.Blue.value}Background{TextStyle.Reset.value}   {FgColor.BrightBlue.value}Foreground{TextStyle.Reset.value} {BgColor.BrightBlue.value}Background{TextStyle.Reset.value}
    Magenta {FgColor.Magenta.value}Foreground{TextStyle.Reset.value} {BgColor.Magenta.value}Background{TextStyle.Reset.value}   {FgColor.BrightMagenta.value}Foreground{TextStyle.Reset.value} {BgColor.BrightMagenta.value}Background{TextStyle.Reset.value}
    Cyan    {FgColor.Cyan.value}Foreground{TextStyle.Reset.value} {BgColor.Cyan.value}Background{TextStyle.Reset.value}   {FgColor.BrightCyan.value}Foreground{TextStyle.Reset.value} {BgColor.BrightCyan.value}Background{TextStyle.Reset.value}
    White   {FgColor.White.value}Foreground{TextStyle.Reset.value} {BgColor.White.value}Background{TextStyle.Reset.value}   {FgColor.BrightWhite.value}Foreground{TextStyle.Reset.value} {BgColor.BrightWhite.value}Background{TextStyle.Reset.value}

    ''')


def showcase_24_bit_colors():
    """Print color chart."""
    from math import sqrt

    print(f"""{TextStyle.Reset.value}24 bit Color Chart
    """)
    size = 16
    nr_rows = int(sqrt(size ** 3))
    nr_cols = nr_rows
    color_bits = int(log2(size))
    color_mask = (0xFF << color_bits) & 0xFF
    # Note [[''*nr_cols]**nr_rows] does not work because all
    # rows will contain references to the same columns.
    matrix = [['' for col in range(nr_cols)] for r in range(nr_rows)]
    for i in range(size ** 3):
        x = i & (nr_cols - 1)
        y = int(i / nr_rows)
        r = (i << color_bits) & color_mask
        g = i & color_mask
        b = (i >> color_bits) & color_mask
        color = AnsiColor(r, g, b)
        matrix[y][x] = f'{color.bg} '
    for row in matrix:
        sys.stdout.write(''.join(row))
        sys.stdout.write(f'{TextStyle.Reset.value}\n')
    sys.stdout.flush()


def showcase_text_styles():
    """Show all implemented text styles."""
    sys.stdout.write(f'''\nText Styles

    Reset            {TextStyle.Reset.value}Reset{TextStyle.Reset.value}
    Bold             {TextStyle.Bold.value}Bold{TextStyle.NormalIntensity.value}             NormalIntensity{TextStyle.Reset.value}
    Faint            {TextStyle.Faint.value}Faint{TextStyle.NormalIntensity.value}            NormalIntensity{TextStyle.Reset.value}
    Italic           {TextStyle.Italic.value}Italic{TextStyle.NoItalic.value}           NoItalic{TextStyle.Reset.value}
    Underline        {TextStyle.Underline.value}Underline{TextStyle.NoUnderline.value}        NoUnderline{TextStyle.Reset.value}
    SlowBlink        {TextStyle.SlowBlink.value}SlowBlink{TextStyle.NoBlink.value}        NoBlink{TextStyle.Reset.value}
    RapidBlink       {TextStyle.RapidBlink.value}RapidBlink{TextStyle.NoBlink.value}       NoBlink{TextStyle.Reset.value}
    Reverse          {TextStyle.Reverse.value}Reverse{TextStyle.NoReverse.value}          NoReverse{TextStyle.Reset.value}
    Hide             {TextStyle.Hide.value}Hide{TextStyle.NoHide.value}             NoHide{TextStyle.Reset.value}
    StrikeOut        {TextStyle.StrikeOut.value}StrikeOut{TextStyle.NoStrikeOut.value}        NoStrikeOut{TextStyle.Reset.value}
    PrimaryFont      {TextStyle.PrimaryFont.value}PrimaryFont{TextStyle.Reset.value}
    AlternateFont1   {TextStyle.AlternateFont1.value}AlternateFont1{TextStyle.Reset.value}
    AlternateFont2   {TextStyle.AlternateFont2.value}AlternateFont2{TextStyle.Reset.value}
    AlternateFont3   {TextStyle.AlternateFont3.value}AlternateFont3{TextStyle.Reset.value}
    AlternateFont4   {TextStyle.AlternateFont4.value}AlternateFont4{TextStyle.Reset.value}
    AlternateFont5   {TextStyle.AlternateFont5.value}AlternateFont5{TextStyle.Reset.value}
    AlternateFont6   {TextStyle.AlternateFont6.value}AlternateFont6{TextStyle.Reset.value}
    AlternateFont7   {TextStyle.AlternateFont7.value}AlternateFont7{TextStyle.Reset.value}
    AlternateFont8   {TextStyle.AlternateFont8.value}AlternateFont8{TextStyle.Reset.value}
    AlternateFont9   {TextStyle.AlternateFont9.value}AlternateFont9{TextStyle.Reset.value}
    GothicFont       {TextStyle.GothicFont.value}GothicFont{TextStyle.Reset.value}
    DoubleUnderline  {TextStyle.DoubleUnderline.value}DoubleUnderline{TextStyle.Reset.value}
    Frame            {TextStyle.Frame.value}Frame{TextStyle.Reset.value}
    Encircle         {TextStyle.Encircle.value}Encircle{TextStyle.Reset.value}
    Overline         {TextStyle.Overline.value}Overline{TextStyle.Reset.value}
    Superscript      {TextStyle.Superscript.value}Superscript{TextStyle.NoSuperSub.value}      NoSuperSub{TextStyle.Reset.value}
    Subscript        {TextStyle.Subscript.value}Subscript{TextStyle.NoSuperSub.value}        NoSuperSub{TextStyle.Reset.value}
    ''')


def showcase_cursor(at: AnsiTerminal):
    """Show cursor movements."""
    from pymisclib.unicodechars import UC

    delay = 0.2
    ts = at.terminal_size

    def show_cursor_position():
        """Print the cursor position in the bottom-left corner."""
        cp = at.cursor_position()
        at.move_cursor_to(ts.column - 7, ts.row)
        sys.stdout.write(f'{cp.column:3d},{cp.row:3d}')
        at.move_cursor_to(cp.column, cp.row)

    sys.stdout.write('\n')
    for l in range(ts.row):
        sys.stdout.write(chr(l + 64) * ts.column)
        sys.stdout.write('\n')
    at.erase_screen()
    at.erase_scrollback_buffer()
    show_cursor_position()
    sleep(delay * 5)

    at.move_cursor_to(1, 1)
    cp1 = at.cursor_position()
    if cp1.column != 1 or cp1.row != 1:
        print('Cursor positions are broken.')
        return
    show_cursor_position()
    sys.stdout.write('\nCursor\n')
    sleep(delay)

    sys.stdout.write('Full line             : 12345678901234567890\n')
    sys.stdout.write('Erase to end of line  : 12345678901234567890')
    at.move_cursor_back(10)
    show_cursor_position()
    sleep(delay * 5)
    at.erase_to_end_of_line()
    sleep(delay)
    sys.stdout.write('\n')
    sys.stdout.write('Erase to start of line: 12345678901234567890')
    at.move_cursor_back(10)
    show_cursor_position()
    sleep(delay * 5)
    at.erase_to_start_of_line()
    sleep(delay)
    sys.stdout.write('\n')
    sys.stdout.write('Erase to entire line  : 12345678901234567890')
    at.move_cursor_back(10)
    show_cursor_position()
    sleep(delay * 5)
    at.erase_line()
    sleep(delay)

    at.move_cursor_down(10)
    show_cursor_position()
    sleep(delay)
    for c in range(40):
        sys.stdout.write(UC.arrows.rightwards_arrow)
        at.move_cursor_back(1)
        at.move_cursor_forward(1)
        show_cursor_position()
        sleep(delay)
    for r in range(5):
        sys.stdout.write(UC.arrows.upwards_arrow)
        at.move_cursor_back(1)
        at.move_cursor_up(1)
        show_cursor_position()
        sleep(delay)
    for c in range(20):
        sys.stdout.write(UC.arrows.leftwards_arrow)
        at.move_cursor_back(1)
        at.move_cursor_back(1)
        show_cursor_position()
        sleep(delay)
    for r in range(5):
        sys.stdout.write(UC.arrows.downwards_arrow)
        at.move_cursor_back(1)
        at.move_cursor_down(1)
        show_cursor_position()
        sleep(delay)
    at.move_cursor_to(1, ts.row)
    print('Done')


def main():
    """Test routine."""
    import os
    os.system('')  # enables ansi escape characters in terminal

    at = AnsiTerminal(full=True, autoflush=True)
    at.erase_screen()

    showcase_cursor(at)
    showcase_3bit_colors()
    showcase_24_bit_colors()
    showcase_text_styles()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# vim ts=4,fileencoding=utf-8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2012-2022, 2023 by Christian Dönges
# SPDXID: SPDXRef-chunkystreamsocket-py
"""A socket used to send and receive message chunks over a TCP connection.

The chunks are assembled back together when receiving.
"""

# pylint: disable=consider-using-assignment-expr

from dataclasses import dataclass
import logging
import socket


@dataclass
class ChunkyStreamSocket:
    """Socket used to send message chunks over a TCP connection."""
    host: str = '127.0.0.1'  # Host name or IP address.
    """Host name or IP address."""
    port: int = 10000
    """Port number."""
    sock: socket.socket | None = None
    """Socket to use."""
    logger: logging.Logger = logging.getLogger(__name__)
    """Logger for diagnostics."""
    debug: bool = False
    """True to enable debug output."""
    log_n_bytes: int = 0   # number of communicated bytes to log
    """Number of communicated bytes to log."""
    family: socket.AddressFamily = socket.AF_INET
    """Address family used by the socket.

    Usually socket.AF_INET [default] or socket.AF_INET6."""
    _backlog_bytes: bytes | None = None
    _connected: bool = False
    _chunk_size: int = 4096
    _server_socket: bool = False  # True for a server socket, False for a client socket.

    def __str__(self) -> str:
        """Human-readable representation of the instance."""
        return f'<{self.__class__.__name__} {self.connection}>'

    @property
    def connected(self) -> bool:
        """Return True if the connection is established."""
        return self._connected

    @property
    def connection(self) -> str:
        """Human-readable description of the underlying connection."""
        if self.sock is None:
            return 'no socket'
        if self._server_socket:
            return f'listening at {self.host}:{self.port}'
        if not self._connected:
            return 'no connection'
        return f'connected to {self.host}:{self.port}'

    @property
    def server_socket(self) -> bool:
        """Is this a server socket (which accepts connections) or a client
            socket, which connects to a server?

        :return: True if server socket, False if client socket.
        """
        return self._server_socket

    def bind_and_listen(self, backlog: int = 5, timeout: float = None):
        """Bind to the host and port to make a server socket.

            :param int backlog: The maximum number of queued connections.
            :param float timeout: The number of seconds after which socket
                 operations will time out. Set to None for a blocking socket.

        """
        self.logger.debug('bind_and_listen(%s:%u)', self.host, self.port)
        if self.sock is None:
            self.sock = socket.socket(self.family, socket.SOCK_STREAM)
        # Set the timeout.
        self.sock.settimeout(timeout)
        # Allow the server socket to re-bind immediately.
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # Bind the socket to the given address and port.
        self.sock.bind((self.host, self.port))
        self._connected = False
        self._server_socket = True
        # Start listening for a connection attempt.
        self.sock.listen(backlog)

    def accept(self) -> tuple[int, tuple[str, int]]:
        """Accept a client connection on the (server) socket.

            :return: The client socket and client address tuple.
            :rtype: tuple[int, tuple[str, int]]
            :raise TimeoutError: No connection accepted withing allowed time.

        """
        try:
            (client_sock, client_addr) = self.sock.accept()  # pylint: disable=no-member
        except socket.timeout as ex:
            raise TimeoutError('accept() timed out') from ex
        self.logger.debug('accept(%s)', client_addr)
        return client_sock, client_addr

    def close(self):
        """Close the socket.

        The underlying OS socket is released and the :py:attr:`sock`
        member will be ``None``.
        """
        self.logger.debug('close() %s', self.connection)
        if self.sock is None:
            return
        try:
            try:
                self.sock.shutdown(socket.SHUT_RDWR)
            except OSError:
                # Ignore errors in shutdown since we close the socket anyways.
                pass
            self.sock.close()
            self.sock = None
        except AttributeError:
            # If the socket has been closed asynchronously, self.socket
            # may be None, resulting in this error.
            pass
        self._client_address = None
        self._connected = False
        self._server_socket = False

    def connect(self, timeout: float|None = None):
        """Connect client socket to the server at host:port.

        :param float timeout: The number of seconds after which socket
            operations will time out. Set to None for a blocking socket.
        :raise BlockingIOError: A non-blocking connection attempt failed because
            establishing the connection takes some time. Try again later.
        :raise ConnectionError: Failed to establish connection.
        :raise InterruptedError: Connection attempt was interrrupted by a signal.
        :raise TimeoutError: Connection attempt timed out.
        """
        if timeout is None:
            self.logger.debug('connect(%s:%u, BLOCKING)', self.host, self.port)
        elif timeout == 0:
            self.logger.debug('connect(%s:%u, NON-BLOCKING)', self.host, self.port)
        else:
            self.logger.debug('connect(%s:%u, TIMEOUT:%f)', self.host, self.port, timeout)
        if self.sock is None:
            self.sock = socket.socket(self.family, socket.SOCK_STREAM)
        # Set the timeout.
        self.sock.settimeout(timeout)
        # Allow socket to re-bind immediately.
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            self.sock.connect((self.host, self.port))
            self._connected = True
        except BlockingIOError as ex:
            self.logger.debug('Connection attempt to %s:%u in progress: %s',
                              self.host, self.port, ex)
            raise ex
        except InterruptedError as ex:
            self.logger.error('Connection attempt to %s:%u interrupted: %s',
                              self.host, self.port, ex)
            raise ex
        except TimeoutError as ex:
            self.logger.error('Connection attempt to %s:%u timed out: %s',
                              self.host, self.port, ex)
            raise ex
        except OSError as ex:
            self.logger.error('Connection attempt to %s:%u failed: %s',
                              self.host, self.port, ex)
            raise ConnectionError('unable to connect') from ex
        self.logger.debug('Connected to %s:%s', self.host, self.port)

    def send(self, msg_bytes: bytes) -> int:
        """Send the bytes.

        :param bytes msg_bytes: The bytes to send. If the receiver is expecting
            a separator, it must be appended to the message by the caller.
        :return: The number of bytes sent.
        :rtype: int
        :raise ConnectionError: sending failed because the connection was broken.
        """
        total_nr_sent = 0
        while total_nr_sent < len(msg_bytes):
            current_nr_sent = 0
            try:
                current_nr_sent = self.sock.send(msg_bytes[total_nr_sent:])
            except ConnectionAbortedError as ex:
                self.logger.debug('Connection ')
            if current_nr_sent == 0:
                self.logger.debug('self.sock.send() failed.')
                self._connected = False
                raise ConnectionError("socket connection broken")
            total_nr_sent = total_nr_sent + current_nr_sent
        if self.log_n_bytes != 0:
            self.logger.debug('--> %s', msg_bytes[:self.log_n_bytes])
        return total_nr_sent

    def recv(self, length: int, timeout: float = None) -> bytes:
        """Receive length bytes from the socket.

        This function handles chunked data (i.e. the data to receive is split
        into multiple packets). If more data than expected is received, it is
        placed into a backlog buffer until the next call to a receive
        function.

        :param int length: Number of bytes to read.
        :param float timeout: Timeout in seconds or `None` for blocking I/O.
        :return: The received bytes.
        :rtype: bytes

        :raise ConnectionError: The connection was broken.
        :raise TimeoutError: Timeout while receiving.
        """
        if length == 0:
            return bytes()
        chunks = []
        nr_received = 0
        # Retrieve the backlog from the previous recv() call and use it as the
        # first chunk.
        if self._backlog_bytes is not None:
            chunks.append(self._backlog_bytes)
            nr_received = len(self._backlog_bytes)
            self._backlog_bytes = None

        # Set the timeout.
        self.sock.settimeout(timeout)

        # Receive bytes until we have enough to satisfy the length requirement.
        while nr_received < length:
            recv_len = min(length - nr_received, self._chunk_size)
            chunk = self.sock.recv(recv_len)
            if self.debug:
                self.logger.debug('socket.recv(%u) := %s', recv_len, chunk)
            if chunk == b'':
                # Connection was closed.
                break
            chunks.append(chunk)
            nr_received = nr_received + len(chunk)

        # Join all chunks into one message.
        msg_bytes = b''.join(chunks)

        # Check if the connection was closed.
        if len(msg_bytes) == 0:
            self.logger.debug('No data received, connection closed.')
            raise ConnectionError("socket connection broken")

        # Cut off the part that is too long.
        if len(msg_bytes) > length:
            self._backlog_bytes = msg_bytes[length:]
            msg_bytes = msg_bytes[:length]

        if self.log_n_bytes != 0:
            self.logger.debug('<-- %s', msg_bytes[:self.log_n_bytes])
        return msg_bytes

    def recv_to_separator(self, separator: bytes) -> bytes:
        """Receive bytes until the given separator is found.

        This function handles chunked data (i.e. the data to receive is
        split into multiple packets). If more data than expected is
        received, it is placed into a backlog buffer until the next call
        to a receive function.

        :param bytes separator: One or more bytes that separate messages in the
            TCP stream.
        :return: The received bytes.
        :rtype: bytes

        :raise ConnectionError: The connection was broken.
        :raise TimeoutError: A timeout occurred while receiving.
        """
        self.logger.debug('recv_to_separator(%s)', separator)
        start_search_index = 0
        chunk = b''
        msg_bytes = b''
        while True:
            if self._backlog_bytes is not None and len(self._backlog_bytes) > 0:
                # The first time around, process the backlog.
                chunk = self._backlog_bytes
                self._backlog_bytes = None
                if self.debug:
                    self.logger.debug('backlog chunk = %s', chunk)
            else:
                chunk = self.sock.recv(self._chunk_size)
                if self.debug:
                    self.logger.debug('socket.recv(%d) := %s', self._chunk_size, chunk)
            if chunk == b'':
                raise ConnectionError("socket connection broken")

            msg_bytes = msg_bytes + chunk
            start_separator_index = msg_bytes.find(separator, start_search_index)
            if start_separator_index > -1:
                # We found the separator at index start_separator_index.
                self._backlog_bytes = msg_bytes[start_separator_index + len(separator):]
                if self.debug:
                    self.logger.debug('Backlog: %u bytes', self._backlog_bytes)
                msg_bytes = msg_bytes[:start_separator_index]
                break
            # The separator could have started in the current chunk but
            # finishes in the next chunk, so we need to search the
            # len(separator) - 1 last bytes of the separator again
            start_search_index = max(0, len(msg_bytes) - (len(separator) - 1))

        if self.log_n_bytes != 0:
            self.logger.debug('<-- %s', msg_bytes[:self.log_n_bytes])
        return msg_bytes

    def __enter__(self):
        """Enter context manager."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit context manager."""
        try:
            if self.sock is not None:
                self.sock.close()
        except Exception:
            pass
        # Propagate any exceptions.
        return False


if __name__ == '__main__':
    import sys
    from threading import Barrier, BrokenBarrierError, Thread
    from time import sleep

    start_barrier = Barrier(2, timeout=5)
    end_barrier = Barrier(3, timeout=60)
    MSG_SEP = b'\0\1\2\1\0'

    messages = [
        b'abcdef',
        b'1234',
        b'a', b'bc', b'de\0', b'\1\2\1\0fghi',
        b'xyzZYX',
        b'',
        b'+++',
    ]

    def server():
        """Test server."""
        logger = logging.getLogger('server')
        logger.info('server starting')
        server_socket = ChunkyStreamSocket(logger=logging.getLogger('server.socket'))
        server_socket.bind_and_listen(timeout=60)
        logger.debug('server_socket = %s', server_socket.sock)
        start_barrier.wait()

        logger.info('server running')
        (accepted_socket, client_addr) = server_socket.accept()
        logger.debug('Accepted client from %s:%u', client_addr[0], client_addr[1])
        cs = ChunkyStreamSocket(sock=accepted_socket,
                                logger=logging.getLogger('server.cs'))
        while True:
            msg = cs.recv_to_separator(MSG_SEP)
            logger.info('MSG: %s', msg)
            if msg == b'+++':
                logger.debug('EOF received')
                break

        logger.info('server closing connection')
        cs.close()
        server_socket.close()
        try:
            end_barrier.wait()
        except BrokenBarrierError:
            logger.error('server() end_barrier broken')

    def client():
        """Test client."""
        logger = logging.getLogger('client')
        logger.info('client starting')
        client_socket = ChunkyStreamSocket(logger=logging.getLogger('client.socket'))
        logger.debug('client_socket = %s', client_socket.sock)
        start_barrier.wait()
        logger.info('client running')
        client_socket.connect()
        for message in messages:
            try:
                client_socket.send(message + MSG_SEP)
            except RuntimeError as e:
                logger.critical('send(%u) failed: %s', len(message + MSG_SEP), e)
                logger.critical('client terminating')
                break
        try:
            end_barrier.wait()
        except BrokenBarrierError:
            logger.error('client() end_barrier broken')
        finally:
            client_socket.close()

    # Log everything to the console.
    main_logger = logging.getLogger()
    main_logger.setLevel(logging.NOTSET)
    ch = logging.StreamHandler()
    ch.setLevel(logging.NOTSET)
    formatter = logging.Formatter('%(asctime)s - %(name)20s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    main_logger.addHandler(ch)

    server_thread = Thread(target=server, args=())
    client_thread = Thread(target=client, args=())
    server_thread.start()
    sleep(1.0)  # give the server time to set up before starting client.
    client_thread.start()
    main_logger.info('Client and server running.')
    try:
        end_barrier.wait(timeout=5)
    except BrokenBarrierError:
        main_logger.error('Barrier broken. Terminating')
    sys.exit(0)



#!/usr/bin/env python3
# vim: fileencoding=utf-8 ts=4
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: © Copyright 2025 by Christian Dönges. All rights reserved.
# SPDXID: SPDXRef-prefixes-py
"""Convert numbers to scaled, prefixed values.

    A unit prefix is a mnemonic that comes before a unit to indicate the
    multiples or fractions of the unit. For example, 0.0001 m becomes 100 µm
    using a metric prefix. 1024 become 1 ki using a binary prefix.

    Example
    -------

    Using a metric (base 10) prefix.

    .. code-block:: python

        for n in [0, 1024, 10000, 10.5678e9, 0.1, .001234, 0.000835, 2.7e-9, ]:
            value, prefix = convert_to_prefix(n, UnitPrefix.Metric)
            print(f'{n} → {value:.2f} {prefix}')
        value, prefix = convert_to_prefix(12349)
        print(f'{value:.0f} {prefix}m')

    Output::

        0 → 0.00
        1024 → 1.02 k
        10000 → 10.00 k
        10567800000.0 → 10.57 G
        0.1 → 100.00 m
        0.001234 → 1.23 m
        0.000835 → 835.00 µ
        2.7e-09 → 2.70 n
        12 km


    Using a binary (base 2) prefix is very similar.

    .. code-block:: python

        for n in [0, 1024, 100000, 9834298724398, 42795864723858438]:
            value, prefix = convert_to_prefix(n, UnitPrefix.Binary)
            print(f'{n} ⇒ {value:.0f} {prefix}')

    Output::

        0 ⇒ 0
        1024 ⇒ 1 Ki
        100000 ⇒ 98 Ki
        9834298724398 ⇒ 9 Ti
        42795864723858438 ⇒ 38 Pi
"""

import enum
import math


@enum.unique
class UnitPrefix(enum.Enum):
    """Prefix type used to indicate fractions or multiples of a unit.

        For example, 1000 m can be expressed as 1 km using the metric prefix
        'kilo' or 'k'.
    """
    Binary = enum.auto()
    """Binary (base 2) prefix."""
    Metric = enum.auto()
    """Metric (base 10) prefix."""


"""A dictionary of prefix definitions.

    Each definition contains a dictionary with the following keys:
        - base: the base of the prefix
        - exponent: the exponent of the prefix
        - min_exponent: the minimum exponent of the prefix
        - max_exponent: the maximum exponent of the prefix
        - prefixes: the actual prefixes, ordered by rising exponent.
            All prefixes are exactly one character long.
        - long_names: a dictionary mapping each prefix to its long name.
            Each long name may be of unique length.
"""
_known_prefixes = {
    UnitPrefix.Binary: {
        'base': 1024,
        'min_exponent': 0,
        'max_exponent': 10,
        'prefixes': [
            '  ',  # 1024^0
            'Ki',  # 1024^1
            'Mi',  # 1024^2
            'Gi',  # 1024^3
            'Ti',  # 1024^4
            'Pi',  # 1024^5
            'Ei',  # 1024^6
            'Zi',  # 1024^7
            'Yi',  # 1024^8
            'Ri',  # 1024^9
            'Qi',  # 1024^10
        ],  # 'prefixes'
        'long_names': {
            '  ': '',
            'Ki': 'kibi',
            'Mi': 'mebi',
            'Gi': 'gibi',
            'Ti': 'tebi',
            'Pi': 'pebi',
            'Ei': 'exbi',
            'Zi': 'zebi',
            'Yi': 'yobi',
            'Ri': 'robi',
            'Qi': 'quebi',
        }  # 'long_names
    },  # Binary
    UnitPrefix.Metric: {
        'base': 1000,
        'min_exponent': -8,
        'max_exponent': 8,
        'prefixes': [
            'y',  # 1000^-8
            'z',  # 1000^-7
            'a',  # 1000^-6
            'f',  # 1000^-5
            'p',  # 1000^-4
            'n',  # 1000^-3
            'µ',  # 1000^-2
            'm',  # 1000^-1
            ' ',  # 1000^0
            'k',  # 1000^1
            'M',  # 1000^2
            'G',  # 1000^3
            'T',  # 1000^4
            'P',  # 1000^5
            'E',  # 1000^6
            'Z',  # 1000^7
            'Y',  # 1000^8
        ],  # prefixes
        'long_names': {
            'y': 'yocto',
            'z': 'zepto',
            'a': 'atto',
            'f': 'femto',
            'p': 'pico',
            'n': 'nano',
            'µ': 'micro',
            'm': 'milli',
            ' ': '',
            'k': 'kilo',
            'M': 'mega',
            'G': 'giga',
            'T': 'terra',
            'P': 'peta',
            'E': 'exa',
            'Z': 'zetta',
            'Y': 'yotta',
        }
    },  # Metric
}  # _known_prefixes


def convert_to_prefix(number: int | float,
                      unit_prefix: UnitPrefix = UnitPrefix.Metric) -> tuple[float, str]:
    """Convert a number to the nearest prefix.

        The nearest prefix is the one with a non-zero digit before the decimal
        separator.

        The valid range of the number depends on the prefix used.

    :param int|float number: The value to convert.
    :param UnitPrefix unit_prefix: The prefix type to use.
    :return: Converted number and prefix.
    :rtype: tuple[float, str]
    """
    pfd = _known_prefixes[unit_prefix]
    if number == 0.0:
        exponent = 0
        scaled = 0
    else:
        exponent = math.floor(math.log(math.fabs(number), pfd['base']))
        scale = math.pow(pfd['base'], exponent)
        scaled = number / scale
    if exponent > pfd['max_exponent']:
        raise ValueError('exponent too large')
    if exponent < pfd['min_exponent']:
        raise ValueError('exponent too small')
    index = exponent - pfd['min_exponent']
    return scaled, pfd['prefixes'][index]



#!/usr/bin/env python3
# vim: fileencoding=utf-8 ts=4
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: © Copyright 2025 by Christian Dönges. All rights reserved.
# SPDXID: SPDXRef-test-utilities-file-birthtime-py

from datetime import datetime, timedelta, timezone
import os
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp, mkstemp
from time import sleep

import pytest
from pymisclib.utilities import OsInfo, file_birthtime


@pytest.fixture
def cleanup():
    """Delete temporary files and directories."""
    paths_to_delete = []
    yield paths_to_delete
    for path in paths_to_delete:
        if os.path.exists(path):
            if os.path.isdir(path):
                rmtree(path)
            else:
                os.remove(path)


def os_dependent_use(path: os.PathLike,
                     tz: timezone|None,
                     fudge: bool) -> datetime|None:
    """Call `file_bithtime()` and interpret the result depending on the OS.

    :param os.PathLike path: Path to a filesystem object (e.g. file or directory.)
    :param timezone|None tz: Timezone used by the filesystem.
    :param bool fudge: If `False`, use only `st_birthtime`. If `True`, try more
        fields.
    :raise AttributeError: Unable to determine creation time without fudging on
        an OS where this is expected to work.
    :return: Creation time of the file or None.
    :rtype: datetime|None
    """
    if fudge:
        return file_birthtime(path, tz,True)
    os_info = OsInfo()
    if os_info.is_windows:
        # Windows has st_birthtime.
        return file_birthtime(path, tz,False)
    if os_info.is_macos or os_info.is_bsd:
        # macOS and some BSDs have st_birthtime.
        return file_birthtime(path, tz,False)
    if os_info.is_linux:
        # Linux does not have st_birthtime.
        with pytest.raises(AttributeError):
            file_birthtime(path, tz,False)
        return None
    # Unknown OS, just try it ...
    return file_birthtime(path, tz,False)


@pytest.mark.parametrize('fudge', [True, False])
def test_file_birthtime(cleanup, fudge):
    tz = datetime.now().astimezone().tzinfo
    max_delta = timedelta(seconds=1)  # this may need to be 2s on Windows with FAT
    ts_start = datetime.now(tz)
    (handle, path) = mkstemp()
    cleanup.append(path)
    ts_created = datetime.now(tz)
    assert ts_start <= ts_created
    os.close(handle)

    ts1 = os_dependent_use(path, tz, fudge)
    if ts1 is None and fudge == False:
        # If there is no st_birthtime, we can't do anything.
        return
    assert ts1 is not None
    # For some reason, on Linux ts1 can be before ts_start.
    # This makes the comparison a little more involved.
    if ts1 <= ts_start:
        assert ts_start - ts1 <= max_delta
    else:
        assert ts1 - ts_start <= max_delta
    if ts1 <= ts_created:
        assert ts_created - ts1 <= max_delta
    else:
        assert ts1 - ts_created <= max_delta
    sleep(2.5)
    ts2 = os_dependent_use(path, tz, fudge)
    assert ts2 is not None
    assert ts1 == ts2
    with open(path, "w+") as f:
        f.write(str(ts1))
    ts3 = file_birthtime(path, tz, fudge)
    if not fudge:
        assert ts1 == ts3
    else:
        assert ts1 <= ts3


@pytest.mark.parametrize('fudge', [True, False])
def test_directory_birthtime(cleanup, fudge):
    max_delta = timedelta(seconds=1)  # this may need to be 2s on Windows with FAT
    tz = datetime.now().astimezone().tzinfo
    ts_start = datetime.now(tz)
    path = mkdtemp()
    cleanup.append(path)
    ts_created = datetime.now(tz)
    assert ts_start <= ts_created
    ts1 = os_dependent_use(path, tz, fudge)
    if ts1 is None and fudge == False:
        # If there is no st_birthtime, we can't do anything.
        return
    assert ts1 is not None
    # For some reason, on Linux ts1 can be before ts_start.
    # This makes the comparison a little more involved.
    if ts1 <= ts_start:
        assert ts_start - ts1 <= max_delta
    else:
        assert ts1 - ts_start <= max_delta
    if ts1 <= ts_created:
        assert ts_created - ts1 <= max_delta
    else:
        assert ts1 - ts_created <= max_delta
    sleep(2.5)
    ts2 = os_dependent_use(path, tz, fudge)
    assert ts2 is not None
    assert ts1 == ts2
    with open(Path(path) / 'file.txt', "w") as f:
        f.write(str(ts1))
    ts3 = os_dependent_use(path, tz, fudge)
    assert ts3 is not None
    if not fudge:
        assert ts1 == ts3
    else:
        assert ts1 <= ts3


@pytest.mark.parametrize('fudge', [True, False])
def test_birthtime_missing_file(fudge: bool):
    """Ensure a missing file raises an exception."""
    with pytest.raises(FileNotFoundError):
        file_birthtime('nonexistent_file', fudge=fudge)
